from functools import singledispatch, wraps
from pandas import DataFrame

import pandas as pd
import numpy as np
import warnings


from pandas.core.groupby import DataFrameGroupBy
from pandas.core.dtypes.inference import is_scalar
from siuba.siu import (
    Symbolic, Call, strip_symbolic, create_sym_call, 
    MetaArg, BinaryOp, _SliceOpIndex, Lazy,
    singledispatch2, pipe_no_args, Pipeable, pipe
    )

from .tidyselect import var_create, var_select, Var

DPLY_FUNCTIONS = (
        # Dply ----
        "group_by", "ungroup", 
        "select", "rename",
        "mutate", "transmute", "filter", "summarize",
        "arrange", "distinct",
        "count", "add_count",
        "head",
        "top_n",
        # Tidy ----
        "spread", "gather",
        "nest", "unnest",
        "expand", "complete",
        "separate", "unite", "extract",
        # Joins ----
        "join", "inner_join", "full_join", "left_join", "right_join", "semi_join", "anti_join",
        # TODO: move to vectors
        "if_else", "case_when",
        "collect", "show_query",
        "tbl",
        )

__all__ = [*DPLY_FUNCTIONS, "Pipeable", "pipe"]


# General TODO ================================================================
# * expressions in group_by
# * n_distinct?
# * separate_rows
# * tally

def install_siu_methods(cls):
    """This function attaches siuba's table verbs on a class, to use as methods.

    """
    func_dict = globals()
    for func_name in DPLY_FUNCTIONS:
        f = func_dict[func_name]

        method_name = "siu_{}".format(func_name)
        setattr(cls, method_name, f)

def install_pd_siu():
    # https://github.com/coursera/pandas-ply/blob/master/pandas_ply/methods.py
    func_dict = globals()
    for func_name in DPLY_FUNCTIONS:
        f = func_dict[func_name]

        method_name = "siu_{}".format(func_name)
        setattr(pd.DataFrame, method_name, f)
        setattr(DataFrameGroupBy, method_name, f)

    DataFrameGroupBy._repr_html_ = _repr_grouped_df_html_
    DataFrameGroupBy.__repr__ = _repr_grouped_df_console_

def _repr_grouped_df_html_(self):
    obj_repr = self.obj._repr_html_()
    
    # user can config pandas not to return html representation, in which case
    # the ipython behavior should fall back to repr
    if obj_repr is None:
        return None

    return "<div><p>(grouped data frame)</p>" + self.obj._repr_html_() + "</div>"

def _repr_grouped_df_console_(self):
    return "(grouped data frame)\n" + repr(self.obj)


def _bounce_groupby(f):
    @wraps(f)
    def wrapper(__data: "pd.DataFrame | DataFrameGroupBy", *args, **kwargs):
        if isinstance(__data, pd.DataFrame):
            return f(__data, *args, **kwargs)

        groupings = __data.grouper.groupings
        group_cols = [ping.name for ping in groupings]

        res = f(__data.obj, *args, **kwargs)

        return res.groupby(group_cols)

    return wrapper


def _regroup(df):
    # try to regroup after an apply, when user kept index (e.g. group_keys = True)
    if len(df.index.names) > 1:
        # handle cases where...
        # 1. grouping with named indices (as_index = True)
        # 2. grouping is level 0 (as_index = False)
        grp_levels = [x for x in df.index.names if x is not None] or [0]

    return df.groupby(level = grp_levels)


def _mutate_cols(__data, args, kwargs):
    from pandas.core.common import apply_if_callable

    result_names = {}          # used as ordered set
    df_tmp = __data.copy()

    for arg in args:

        # case 1: a simple, existing name is a no-op ----
        simple_name = simple_varname(arg)
        if simple_name is not None and simple_name in df_tmp.columns:
            result_names[simple_name] = True
            continue

        # case 2: across ----
        # TODO: make robust. validate input. validate output (e.g. shape).
        res_arg = arg(df_tmp)

        if not isinstance(res_arg, pd.DataFrame):
            raise NotImplementedError("Only across() can be used as positional argument.")

        # unpack result
        is_scalar = len(res_arg) == 1

        for col_name, col_ser in res_arg.items():
            # need to put on the frame so subsequent args, kwargs can use
            if is_scalar:
                df_tmp.loc[:, col_name] = col_ser.iloc[0]
            else:
                df_tmp.loc[:, col_name] = col_ser.array

            result_names[col_name] = True

    for col_name, expr in kwargs.items():
        # this is exactly what DataFrame.assign does
        df_tmp[col_name] = apply_if_callable(expr, df_tmp)
        result_names[col_name] = True

    return list(result_names), df_tmp


def _make_groupby_safe(gdf):
    return gdf.obj.groupby(gdf.grouper, group_keys=False, dropna=False)


MSG_TYPE_ERROR = "The first argument to {func} must be one of: {types}"

def raise_type_error(f):
    raise TypeError(MSG_TYPE_ERROR.format(
                func = f.__name__,
                types = ", ".join(map(str, f.registry.keys()))
                ))


def simple_varname(call):
    if isinstance(call, str):
        return call

    # check for expr like _.some_var or _["some_var"]
    if (isinstance(call, Call)
        and call.func in {"__getitem__", "__getattr__"}
        and isinstance(call.args[0], MetaArg)
        and isinstance(call.args[1], (str, _SliceOpIndex))
        ):
        # return variable name
        name = call.args[1]

        if isinstance(name, str):
            return name
        elif isinstance(name, _SliceOpIndex):
            return name.args[0]

    return None


def ordered_union(*args):
    out = {}
    for arg in args:
        out.update({el: True for el in arg})
    return tuple(out)


# Collect and show_query =========

@pipe_no_args
@singledispatch2((DataFrame, DataFrameGroupBy))
def collect(__data, *args, **kwargs):
    """Retrieve data as a local DataFrame.
    
    """
    # simply return DataFrame, since requires no execution
    return __data


@pipe_no_args
@singledispatch2((DataFrame, DataFrameGroupBy))
def show_query(__data, simplify = False):
    """Print the details of a query.

    Parameters
    ----------
    __data:
        A DataFrame of siuba.sql.LazyTbl.
    simplify:
        Whether to attempt to simplify the query.
    **kwargs:
        Additional arguments passed to specific implementations.

    """
    print("No query to show for a DataFrame")
    return __data

# Mutate ======================================================================


@singledispatch2(pd.DataFrame)
def mutate(__data, *args, **kwargs):
    """Assign new variables to a DataFrame, while keeping existing ones.

    Parameters
    ----------
    __data: pd.DataFrame
    **kwargs:
        new_col_name=value pairs, where value can be a function taking a singledispatch2
        argument for the data being operated on.

    See Also
    --------
    transmute : Returns a DataFrame with only the newly created columns.

    Examples
    --------
    >>> from siuba import _, mutate, head
    >>> from siuba.data import cars
    >>> cars >> mutate(cyl2 = _.cyl * 2, cyl4 = _.cyl2 * 2) >> head(2)
       cyl   mpg   hp  cyl2  cyl4
    0    6  21.0  110    12    24
    1    6  21.0  110    12    24
        
    """

    new_names, df_res = _mutate_cols(__data, args, kwargs)
    return df_res


@mutate.register(DataFrameGroupBy)
def _mutate(__data, *args, **kwargs):
    out = __data.obj.copy()
    groupings = {ping.name: ping for ping in __data.grouper.groupings}

    f_transmute = transmute.dispatch(pd.DataFrame)

    df = _make_groupby_safe(__data).apply(lambda d: f_transmute(d, *args, **kwargs))

    for varname, ser in df.items():
        if varname in groupings:
            groupings[varname] = varname

        out[varname] = ser

    return out.groupby(list(groupings.values()))


# Group By ====================================================================

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def group_by(__data, *args, add = False, **kwargs):
    """Return a grouped DataFrame, using columns or expressions to define groups.

    Any operations (e.g. summarize, mutate, filter) performed on grouped data
    will be performed "by group". Use `ungroup()` to remove the groupings.

    Parameters
    ----------
    __data:
        The data being grouped.
    *args:
        Lazy expressions used to select the grouping columns. Currently, each
        arg must refer to a single columns (e.g. _.cyl, _.mpg).
    add: bool
        If the data is already grouped, whether to add these groupings on top of those.
    **kwargs:
        Keyword arguments define new columns used to group the data.


    Examples
    --------

    >>> from siuba import _, group_by, summarize, filter, mutate, head
    >>> from siuba.data import cars

    >>> by_cyl = cars >> group_by(_.cyl)

    >>> by_cyl >> summarize(max_mpg = _.mpg.max(), max_hp = _.hp.max())
       cyl  max_mpg  max_hp
    0    4     33.9     113
    1    6     21.4     175
    2    8     19.2     335

    >>> by_cyl >> filter(_.mpg == _.mpg.max())
    (grouped data frame)
        cyl   mpg   hp
    3     6  21.4  110
    19    4  33.9   65
    24    8  19.2  175

    >>> cars >> group_by(cyl2 = _.cyl + 1) >> head(2)
    (grouped data frame)
       cyl   mpg   hp  cyl2
    0    6  21.0  110     7
    1    6  21.0  110     7

    Note that creating the new grouping column is always performed on ungrouped data.
    Use an explicit mutate on the grouped data perform the operation within groups.

    For example, the code below calls pd.cut on the mpg column, within each cyl group.

    >>> from siuba.siu import call
    >>> (cars
    ...     >> group_by(_.cyl)
    ...     >> mutate(mpg_bin = call(pd.cut, _.mpg, 3))
    ...     >> group_by(_.mpg_bin, add=True)
    ...     >> head(2)
    ... )
    (grouped data frame)
       cyl   mpg   hp       mpg_bin
    0    6  21.0  110  (20.2, 21.4]
    1    6  21.0  110  (20.2, 21.4]
    
    """
    
    if isinstance(__data, DataFrameGroupBy):
        tmp_df = __data.obj.copy()
    else:
        tmp_df = __data.copy()

    # TODO: super inefficient, since it makes multiple copies of data
    #       need way to get the by_vars and apply (grouped) computation
    computed = transmute(tmp_df, *args, **kwargs)
    by_vars = list(computed.columns)

    for k in by_vars:
        tmp_df[k] = computed[k]

    if isinstance(__data, DataFrameGroupBy) and add:
        groupings = {el.name: el for el in __data.grouper.groupings}

        for varname in by_vars:
            # ensures group levels are recalculated if varname was in transmute
            groupings[varname] = varname

        return tmp_df.groupby(list(groupings.values()), dropna=False, group_keys=True)

    return tmp_df.groupby(by = by_vars, dropna=False, group_keys=True)


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def ungroup(__data):
    """Return an ungrouped DataFrame.

    Parameters
    ----------
    __data:
        The data being ungrouped.

    Examples
    --------
    >>> from siuba import _, group_by, ungroup
    >>> from siuba.data import cars

    >>> g_cyl = cars.groupby("cyl")
    >>> res1 = ungroup(g_cyl)

    >>> res2 = cars >> group_by(_.cyl) >> ungroup()
    """
    # TODO: can we somehow just restore the original df used to construct
    #       the groupby?
    if isinstance(__data, pd.DataFrame):
        return __data
    elif isinstance(__data, DataFrameGroupBy):
        return __data.obj
    else:
        raise TypeError(f"Unsupported type {type(__data)}")



# Filter ======================================================================

@singledispatch2(pd.DataFrame)
def filter(__data, *args):
    """Keep rows where conditions are true.

    Parameters
    ----------
    __data:
        The data being filtered.
    *args:
        conditions that must be met to keep a column. 

    Examples
    --------

    >>> from siuba import _, filter
    >>> from siuba.data import cars
    
    Keep rows where cyl is 4 *and* mpg is less than 25.

    >>> cars >> filter(_.cyl ==  4, _.mpg < 22) 
        cyl   mpg   hp
    20    4  21.5   97
    31    4  21.4  109

    Use `|` to represent an OR condition. For example, the code below keeps
    rows where hp is over 250 *or* mpg is over 32.

    >>> cars >> filter((_.hp > 300) | (_.mpg > 32))
        cyl   mpg   hp
    17    4  32.4   66
    19    4  33.9   65
    30    8  15.0  335

    """
    crnt_indx = True
    for arg in args:
        res = arg(__data) if callable(arg) else arg

        if isinstance(res, pd.DataFrame):
            crnt_indx &= res.all(axis=1)
        elif isinstance(res, pd.Series):
            crnt_indx &= res
        else:
            crnt_indx &= res

    # use loc or iloc to subset, depending on crnt_indx ----
    # the main issue here is that loc can't remove all rows using a slice
    # and iloc can't use a boolean series
    if isinstance(crnt_indx, bool) or isinstance(crnt_indx, np.bool_):
        # iloc can do slice, but not a bool series
        result = __data.iloc[slice(None) if crnt_indx else slice(0),:]
    else:
        result = __data.loc[crnt_indx,:]

    return result


@filter.register(DataFrameGroupBy)
def _filter(__data, *args):
    groupings = __data.grouper.groupings
    df_filter = filter.registry[pd.DataFrame]

    df = __data.apply(df_filter, *args)

    # will drop all but original index, then sort to get original order
    group_by_lvls = list(range(df.index.nlevels - 1))
    ordered = df.reset_index(group_by_lvls, drop = True).sort_index()

    group_cols = [ping.name for ping in groupings]
    return ordered.groupby(group_cols)


# Summarize ===================================================================


@singledispatch2(DataFrame)
def summarize(__data, *args, **kwargs):
    """Assign variables that are single number summaries of a DataFrame.

    Grouped DataFrames will produce one row for each group. Otherwise, summarize
    produces a DataFrame with a single row.

    Parameters
    ----------
    __data: a DataFrame
        The data being summarized.
    **kwargs:
        new_col_name=value pairs, where value can be a function taking
        a single argument for the data being operated on.


    Examples
    --------
    >>> from siuba import _, group_by, summarize
    >>> from siuba.data import cars

    >>> cars >> summarize(avg = _.mpg.mean(), n = _.shape[0])
             avg   n
    0  20.090625  32

    >>> g_cyl = cars >> group_by(_.cyl)
    >>> g_cyl >> summarize(min = _.mpg.min())
       cyl   min
    0    4  21.4
    1    6  17.8
    2    8  10.4

    >>> g_cyl >> summarize(mpg_std_err = _.mpg.std() / _.shape[0]**.5)
       cyl  mpg_std_err
    0    4     1.359764
    1    6     0.549397
    2    8     0.684202
        
    """
    results = {}
    
    for ii, expr in enumerate(args):
        if not callable(expr):
            raise TypeError(
                "Unnamed arguments to summarize must be callable, but argument number "
                f"{ii} was type: {type(expr)}"
            )

        res = expr(__data)
        if isinstance(res, DataFrame):
            if len(res) != 1:
                raise ValueError(
                    f"Summarize argument `{ii}` returned a DataFrame with {len(res)} rows."
                    " Result must only be a single row."
                )

            for col_name in res.columns:
                results[col_name] = res[col_name].array
        else:
            raise ValueError(
                "Unnamed arguments to summarize must return a DataFrame, but argument "
                f"`{ii} returned type: {type(expr)}"
            )



    for k, v in kwargs.items():
        # TODO: raise error if a named expression returns a DataFrame
        res = v(__data) if callable(v) else v

        if is_scalar(res) or len(res) == 1:
            # keep result, but use underlying array to avoid crazy index issues
            # on DataFrame construction (#138)
            results[k] = res.array if isinstance(res, pd.Series) else res

        else:
            raise ValueError(
                f"Summarize argument `{k}` must return result of length 1 or a scalar.\n\n"
                f"Result type: {type(res)}\n"
                f"Result length: {len(res)}"
            )
        
    # must pass index, or raises error when using all scalar values
    return DataFrame(results, index = [0])

    
@summarize.register(DataFrameGroupBy)
def _summarize(__data, *args, **kwargs):
    if __data.dropna or not __data.group_keys:
        warnings.warn(
            f"Grouped data passed to summarize must have dropna=False and group_keys=True."
            " Regrouping with these arguments set."
        )

        if __data.grouper.dropna:
            # will need to recalculate groupings, otherwise it ignores dropna
            group_cols = [ping.name for ping in __data.grouper.groupings]
        else:
            group_cols = __data.grouper.groupings
        __data = __data.obj.groupby(group_cols, dropna=False, group_keys=True)

    df_summarize = summarize.registry[pd.DataFrame]

    df = __data.apply(df_summarize, *args, **kwargs)
        
    group_by_lvls = list(range(df.index.nlevels - 1))
    out = df.reset_index(group_by_lvls)
    out.index = pd.RangeIndex(df.shape[0])

    return out



# Transmute ===================================================================

@singledispatch2(DataFrame)
def transmute(__data, *args, **kwargs):
    """Assign new columns to a DataFrame, while dropping previous columns.

    Parameters
    ----------
    __data:
        The input data.
    **kwargs:
        Each keyword argument is the name of a new column, and an expression.

    See Also
    --------
    mutate : Assign new columns, or modify existing ones.

    Examples
    --------

    >>> from siuba import _, transmute, mutate, head
    >>> from siuba.data import cars

    Notice that transmute results in a table with only the new column:

    >>> cars >> transmute(cyl2 = _.cyl + 1) >> head(2)
       cyl2
    0     7
    1     7

    By contrast, mutate adds the new column to the end of the table:

    >>> cars >>  mutate(cyl2 = _.cyl + 1) >> head(2)
       cyl   mpg   hp  cyl2
    0    6  21.0  110     7
    1    6  21.0  110     7
    

    """
    arg_vars = list(map(simple_varname, args))

    col_names, df_res = _mutate_cols(__data, args, kwargs)
    return df_res[col_names]


@transmute.register(DataFrameGroupBy)
def _transmute(__data, *args, **kwargs):
    groupings = {ping.name: ping for ping in __data.grouper.groupings}

    f_transmute = transmute.dispatch(pd.DataFrame)

    df = _make_groupby_safe(__data).apply(lambda d: f_transmute(d, *args, **kwargs))

    
    for varname in reversed(list(groupings)):
        if varname in df.columns:
            groupings[varname] = varname
        else:
            df.insert(0, varname, __data.obj[varname])

    return df.groupby(list(groupings.values()))



# Select ======================================================================

def _insert_missing_groups(dst, orig, missing_groups):
    if missing_groups:
        warnings.warn(f"Adding missing grouping variables: {missing_groups}")

        for ii, colname in enumerate(missing_groups):
            dst.insert(ii, colname, orig[colname])


def _select_group_renames(selection: dict, group_cols):
    """Returns a 2-tuple: groups missing in the select, new group keys."""
    renamed = {k: v for k,v in selection.items() if v is not None}

    sel_groups = [
        renamed[colname] or colname for colname in group_cols if colname in renamed
    ]
    missing_groups = [colname for colname in group_cols if colname not in selection]

    return missing_groups, (*missing_groups, *sel_groups)


@singledispatch2(DataFrame)
def select(__data, *args, **kwargs):
    """Select columns of a table to keep or drop (and optionally rename).

    Parameters
    ----------
    __data:
        The input table.
    *args: 
        An expression specifying columns to keep or drop. 
    **kwargs:
        Not implemented.

    Examples
    --------
    >>> from siuba import _, select
    >>> from siuba.data import cars

    >>> small_cars = cars.head(1)
    >>> small_cars
       cyl   mpg   hp
    0    6  21.0  110

    You can refer to columns by name or position.

    >>> small_cars >> select(_.cyl, _[2])
       cyl   hp
    0    6  110

    Use a `~` sign to exclude a column.

    >>> small_cars >> select(~_.cyl)
        mpg   hp
    0  21.0  110

    You can use any methods you'd find on the .columns.str accessor:

    >>> small_cars.columns.str.contains("p")
    array([False,  True,  True])

    >>> small_cars >> select(_.contains("p"))
        mpg   hp
    0  21.0  110

    Use a slice to select a range of columns:

    >>> small_cars >> select(_[0:2])
       cyl   mpg
    0    6  21.0

    Multiple expressions can be combined using _[a, b, c] syntax. This is useful
    for dropping a complex set of matches.

    >>> small_cars >> select(~_[_.startswith("c"), -1])
        mpg
    0  21.0

    """

    if kwargs:
        raise NotImplementedError(
                "Using kwargs in select not currently supported. "
                "Use _.newname == _.oldname instead"
                )
    var_list = var_create(*args)

    od = var_select(__data.columns, *var_list, data=__data)

    to_rename = {k: v for k,v in od.items() if v is not None}

    return __data[list(od)].rename(columns = to_rename)
    

@select.register(DataFrameGroupBy)
def _select(__data, *args, **kwargs):
    # tidyselect
    var_list = var_create(*args)
    od = var_select(__data.obj.columns, *var_list)

    group_cols = [ping.name for ping in __data.grouper.groupings]

    res = select(__data.obj, *args, **kwargs)

    missing_groups, group_keys = _select_group_renames(od, group_cols)
    _insert_missing_groups(res, __data.obj, missing_groups)

    return res.groupby(list(group_keys))


# Rename ======================================================================

@singledispatch2(DataFrame)
def rename(__data, **kwargs):
    """Rename columns of a table.

    Parameters
    ----------
    __data:
        The input table.
    **kwargs:
        Keyword arguments of the form new_name = _.old_name, or new_name = "old_name".

    Examples
    --------

    >>> import pandas as pd
    >>> from siuba import _, rename, select

    >>> df = pd.DataFrame({"zzz": [1], "b": [2]})
    >>> df >> rename(a = _.zzz)
       a  b
    0  1  2

    Note that this is equivalent to this select code:

    >>> df >> select(_.a == _.zzz, _.b)
       a  b
    0  1  2

    """
    # TODO: allow names with spaces, etc..
    col_names = {simple_varname(v):k for k,v in kwargs.items()}
    if None in col_names:
        raise ValueError("Rename needs column name (e.g. 'a' or _.a), but received %s"%col_names[None])

    return __data.rename(columns  = col_names)


@rename.register(DataFrameGroupBy)
def _rename(__data, **kwargs):
    col_names = {simple_varname(v):k for k,v in kwargs.items()}
    group_cols = [ping.name for ping in __data.grouper.groupings]

    res = rename(__data.obj, **kwargs)

    missing_groups, group_keys = _select_group_renames(col_names, group_cols)

    return res.groupby(list(group_keys))


# Arrange =====================================================================

def _call_strip_ascending(f):
    if isinstance(f, Symbolic):
        f = strip_symbolic(f)

    if isinstance(f, Call) and f.func == "__neg__":
        return f.args[0], False

    return f, True

@singledispatch2(DataFrame)
def arrange(__data, *args):
    """Re-order the rows of a DataFrame using the values of specified columns.

    Parameters
    ----------
    __data:
        The input table.
    *args:
        Columns or expressions used to sort the rows.

    Examples
    --------

    >>> import pandas as pd
    >>> from siuba import _, arrange, mutate

    >>> df = pd.DataFrame({"x": [2, 1, 1], "y": ["aa", "b", "aa"]})
    >>> df
       x   y
    0  2  aa
    1  1   b
    2  1  aa

    Arrange sorts on the first argument, then the second, etc..

    >>> df >> arrange(_.x, _.y)
       x   y
    2  1  aa
    1  1   b
    0  2  aa

    Use a minus sign (`-`) to sort is descending order.

    >>> df >> arrange(-_.x)
       x   y
    0  2  aa
    1  1   b
    2  1  aa

    Note that arrange can sort on complex expressions:

    >>> df >> arrange(-_.y.str.len())
       x   y
    0  2  aa
    2  1  aa
    1  1   b

    The case above is equivalent to running a mutate before arrange:

    >>> df >> mutate(res = -_.y.str.len()) >> arrange(_.res)
       x   y  res
    0  2  aa   -2
    2  1  aa   -2
    1  1   b   -1

    """
    # TODO:
    #   - add arguments to pass to sort_values (e.g. ascending, kind)
    # 
    # basically need some (1) select behavior, (2) mutate-like behavior
    # df.sort_values is the obvious candidate, but only takes names, not expressions
    # to work around this, we make a shallow copy of data, and add sorting columns
    # then drop them at the end
    # 
    # sort order is determined by using a unary w/ Call e.g. -_.repo

    df = __data.copy(deep = False)
    n_cols = len(df.columns)
    n_args = len(args)

    #kwargs = {n_cols + ii: arg for ii,arg in enumerate(args)}

    # TODO: more careful handling of arg types (true across library :/ )..
    tmp_cols = []
    sort_cols = []
    ascending = []
    for ii, arg in enumerate(args):
        f, asc = _call_strip_ascending(arg)

        ascending.append(asc)

        col = simple_varname(f)
        if col is not None:
            sort_cols.append(col)
        else:
            # TODO: could screw up if user has columns names that are ints...
            sort_cols.append(n_cols + ii)
            tmp_cols.append(n_cols + ii)

            res = f(df)

            if isinstance(res, pd.DataFrame):
                raise NotImplementedError(
                    f"`arrange()` expression {ii} of {len(args)} returned a "
                    "DataFrame, which is currently unsupported."
                )

            df[n_cols + ii] = res


    return df.sort_values(by = sort_cols, kind = "mergesort", ascending = ascending) \
             .drop(tmp_cols, axis = 1)


@arrange.register(DataFrameGroupBy)
def _arrange(__data, *args):
    for arg in args:
        f, desc = _call_strip_ascending(arg)
        if not simple_varname(f):
            raise NotImplementedError(
                    "Arrange over DataFrameGroupBy only supports simple "
                    "column names, not expressions"
                    )

    df_sorted = arrange(__data.obj, *args)

    group_cols = [ping.name for ping in __data.grouper.groupings]
    return df_sorted.groupby(group_cols)




# Distinct ====================================================================


@singledispatch2(DataFrame)
def distinct(__data, *args, _keep_all = False, **kwargs):
    """Keep only distinct (unique) rows from a table.

    Parameters
    ----------
    __data:
        The input data.
    *args:
        Columns to use when determining which rows are unique.
    _keep_all:
        Whether to keep all columns of the original data, not just *args.
    **kwargs:
        If specified, arguments passed to the verb mutate(), and then being used
        in distinct().

    See Also
    --------
    count : keep distinct rows, and count their number of observations.

    Examples
    --------
    >>> from siuba import _, distinct, select
    >>> from siuba.data import penguins

    >>> penguins >> distinct(_.species, _.island)
         species     island
    0     Adelie  Torgersen
    1     Adelie     Biscoe
    2     Adelie      Dream
    3     Gentoo     Biscoe
    4  Chinstrap      Dream

    Use _keep_all=True, to keep all columns in each distinct row. This lets you
    peak at the values of the first unique row.

    >>> small_penguins = penguins >> select(_[:4])
    >>> small_penguins >> distinct(_.species, _keep_all = True)
         species     island  bill_length_mm  bill_depth_mm
    0     Adelie  Torgersen            39.1           18.7
    1     Gentoo     Biscoe            46.1           13.2
    2  Chinstrap      Dream            46.5           17.9
    """

    if not (args or kwargs):
        return __data.drop_duplicates().reset_index(drop=True)

    new_names, df_res = _mutate_cols(__data, args, kwargs)
    tmp_data = df_res.drop_duplicates(new_names).reset_index(drop=True)

    if not _keep_all:
        return tmp_data[new_names]

    return tmp_data


@distinct.register(DataFrameGroupBy)
def _distinct(__data, *args, _keep_all = False, **kwargs):

    group_names = [ping.name for ping in __data.grouper.groupings]


    f_distinct = distinct.dispatch(type(__data.obj))

    tmp_data = (__data
        .apply(f_distinct, *args, _keep_all=_keep_all, **kwargs)
    )

    index_keys = tmp_data.index.names[:-1]
    keys_to_drop = [k for k in index_keys if k in tmp_data.columns]
    keys_to_keep = [k for k in index_keys if k not in tmp_data.columns]

    final = tmp_data.reset_index(keys_to_drop, drop=True).reset_index(keys_to_keep)

    return final.groupby(group_names)


# if_else, case_when ==========================================================

# TODO: move to vector.py
@singledispatch
def if_else(condition, true, false):
    """
    Parameters
    ----------
    condition:
        Logical vector (or lazy expression).
    true:
        Values to be used when condition is True.
    false:
        Values to be used when condition is False.

    See Also
    --------
    case_when : Generalized if_else, for handling many cases.
        
    Examples
    --------
    >>> ser1 = pd.Series([1,2,3])
    >>> if_else(ser1 > 2, np.nan, ser1)
    0    1.0
    1    2.0
    2    NaN
    dtype: float64

    >>> from siuba import _
    >>> f = if_else(_ < 2, _, 2)
    >>> f(ser1)
    0    1
    1    2
    2    2
    dtype: int64

    >>> import numpy as np
    >>> ser2 = pd.Series(['NA', 'a', 'b'])
    >>> if_else(ser2 == 'NA', np.nan, ser2)
    0    NaN
    1      a
    2      b
    dtype: object

    """
    raise_type_error(condition)

@if_else.register(Call)
@if_else.register(Symbolic)
def _if_else(condition, true, false):
    return create_sym_call(if_else, condition, true, false)

@if_else.register(pd.Series)
def _if_else(condition, true, false):
    result = np.where(condition.fillna(False), true, false)

    return pd.Series(result)


# case_when ----------------
# note that here, we don't use @Pipeable.add_to_dispatcher.
# because case_when takes a dictionary of cases, we need to wrap cases into
# a Call, so that it can be handled by call tree visitors, etc..
# TODO: evaluate this non-table verb approach
from siuba.siu import DictCall

def _val_call(call, data, n, indx = None):
    if not callable(call):
        return call

    arr = call(data)
    if arr.shape != (n,):
        raise ValueError("Expected call to return array of shape {}"
                         "but it returned shape {}".format(n, arr.shape))

    return arr[indx] if indx is not None else arr


@singledispatch2((pd.DataFrame,pd.Series))
def case_when(__data, cases: dict):
    """Generalized, vectorized if statement.

    Parameters
    ----------
    __data:
        The input data.
    cases: dict
        A mapping of condition : value.

    See Also
    --------
    if_else : Handles the special case of two conditions.
        
    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, case_when

    >>> df = pd.DataFrame({"x": [1, 2, 3]})
    >>> case_when(df, {_.x == 1: "one", _.x == 2: "two"})
    0     one
    1     two
    2    None
    dtype: object

    >>> df >> case_when({_.x == 1: "one", _.x == 2: "two"})
    0     one
    1     two
    2    None
    dtype: object

    >>> df >> case_when({_.x == 1: "one", _.x == 2: "two", True: "other"})
    0      one
    1      two
    2    other
    dtype: object


    """
    if isinstance(cases, Call):
        cases = cases(__data)
    # TODO: handle when receive list of (k,v) pairs for py < 3.5 compat?

    stripped_cases = {strip_symbolic(k): strip_symbolic(v) for k,v in cases.items()}
    n = len(__data)
    out = np.repeat(None, n)
    for k, v in reversed(list(stripped_cases.items())):
        if callable(k):
            result = _val_call(k, __data, n)
            indx = np.where(result)[0]

            val_res = _val_call(v, __data, n, indx)
            out[indx] = val_res
        elif k:
            # e.g. k is just True, etc..
            val_res = _val_call(v, __data, n)
            out[:] = val_res

    # by recreating an array, attempts to cast as best dtype
    return pd.Series(list(out))

@case_when.register(Symbolic)
@case_when.register(Call)
def _case_when(__data, cases):
    if not isinstance(cases, dict):
        raise Exception("Cases must be a dictionary")
    dict_entries = dict((strip_symbolic(k), strip_symbolic(v)) for k,v in cases.items())
    cases_arg = Lazy(DictCall("__call__", dict, dict_entries))
    return create_sym_call(case_when, __data, cases_arg)




# Count =======================================================================

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def count(__data, *args, wt = None, sort = False, name=None, **kwargs):
    """Summarize data with the number of rows for each grouping of data.

    Parameters
    ----------
    __data:
        A DataFrame.
    *args:
        The names of columns to be used for grouping. Passed to group_by.
    wt:
        The name of a column to use as a weighted for each row.
    sort:
        Whether to sort the results in descending order.
    **kwargs:
        Creates a new named column, and uses for grouping. Passed to group_by.

    Examples
    --------

    >>> from siuba import _, count, group_by, summarize, arrange
    >>> from siuba.data import mtcars

    >>> count(mtcars, _.cyl, high_mpg = _.mpg > 30)
       cyl  high_mpg   n
    0    4     False   7
    1    4      True   4
    2    6     False   7
    3    8     False  14

    Use sort to order results by number of observations (in descending order).

    >>> count(mtcars, _.cyl, sort=True)
       cyl   n
    0    8  14
    1    4  11
    2    6   7

    count is equivalent to doing a grouped summarize:

    >>> mtcars >> group_by(_.cyl) >> summarize(n = _.shape[0]) >> arrange(-_.n)
       cyl   n
    2    8  14
    0    4  11
    1    6   7


    """
    no_grouping_vars = not args and not kwargs and isinstance(__data, pd.DataFrame)

    if wt is None:
        if no_grouping_vars: 
            # no groups, just use number of rows
            counts = pd.DataFrame({'tmp': [__data.shape[0]]})
        else:
            # tally rows for each group
            counts = group_by(__data, *args, add = True, **kwargs).size().reset_index()
    else:
        wt_col = simple_varname(wt)
        if wt_col is None:
            raise Exception("wt argument has to be simple column name")

        if no_grouping_vars:
            # no groups, sum weights
            counts = pd.DataFrame({'tmp': [__data[wt_col].sum()]})
        else:
            # do weighted tally
            counts = group_by(__data, *args, add = True, **kwargs)[wt_col].sum().reset_index()


    # count col named, n. If that col already exists, add more "n"s...
    out_col = _check_name(name, set(counts.columns))

    # rename the tally column to correct name
    counts.rename(columns = {counts.columns[-1]: out_col}, inplace = True)

    if sort:
        return counts.sort_values(out_col, ascending = False).reset_index(drop = True)

    return counts


def _check_name(name, columns):
    if name is None:
        name = "n"
        while name in columns:
            name = name + "n"

    elif name != "n" and name in columns:
        raise ValueError(
            f"Column name `{name}` specified for count name, but is already present in data."
        )
    
    elif not isinstance(name, str):
        raise TypeError("`name` must be a single string.")

    return name
        

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def add_count(__data, *args, wt = None, sort = False, name = None, **kwargs):
    """Add a column that is the number of observations for each grouping of data.

    Note that this function is similar to count(), but does not aggregate. It's
    useful combined with filter().

    Parameters
    ----------
    __data:
        A DataFrame.
    *args:
        The names of columns to be used for grouping. Passed to group_by.
    wt:
        The name of a column to use as a weighted for each row.
    sort:
        Whether to sort the results in descending order.
    **kwargs:
        Creates a new named column, and uses for grouping. Passed to group_by.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, add_count, group_by, ungroup, mutate
    >>> from siuba.data import mtcars

    >>> df = pd.DataFrame({"x": ["a", "a", "b"], "y": [1, 2, 3]})
    >>> df >> add_count(_.x)
       x  y  n
    0  a  1  2
    1  a  2  2
    2  b  3  1

    This is useful if you want to see data associated with some count:

    >>> df >> add_count(_.x) >> filter(_.n == 1)
       x  y  n
    2  b  3  1

    Note that add_count is equivalent to a grouped mutate:

    >>> df >> group_by(_.x) >> mutate(n = _.shape[0]) >> ungroup()
       x  y  n
    0  a  1  2
    1  a  2  2
    2  b  3  1


    """

    no_grouping_vars = not args and not kwargs and isinstance(__data, pd.DataFrame)

    if no_grouping_vars:
        out = __data
    else:
        out = group_by(__data, *args, add=True, **kwargs)

    var_names = ungroup(out).columns
    name = _check_name(name, set(var_names))

    if wt is None:
        if no_grouping_vars: 
            # no groups, just use number of rows
            counts = __data.copy()
            counts[name] = counts.shape[0]
        else:
            # note that it's easy to transform tally using single grouped column, so
            # we arbitrarily grab the first column..
            counts = out.obj.copy()
            counts[name] = out[var_names[0]].transform("size")

    else:
        wt_col = simple_varname(wt)
        if wt_col is None:
            raise Exception("wt argument has to be simple column name")

        if no_grouping_vars:
            # no groups, sum weights
            counts = __data.copy()
            counts[name] = counts[wt_col].sum()
        else:
            # TODO: should flip topmost if/else so grouped code is together
            # do weighted tally
            counts = out.obj.copy()
            counts[name] = out[wt_col].transform("sum")

    if sort:
        return counts.sort_values(out_col, ascending = False)

    return counts

    


# Tally =======================================================================

# Nest ========================================================================

def _fast_split_df(g_df):
    """
    Note
    ----

    splitting does not scale well to many groups (e.g. 50000+). This is due
    to pandas' (1) use of indexes, (2) some hard coded actions when subsetting.
    We are currently working on a fix, so that when people aren't using indexes,
    nesting will be much faster.

    see https://github.com/machow/siuba/issues/184
    """

    # TODO (#184): speed up when user doesn't need an index
    # right now, this is essentially a copy of
    # pandas.core.groupby.ops.DataSplitter.__iter__
    from pandas._libs import lib
    splitter = g_df.grouper._get_splitter(g_df.obj)

    starts, ends = lib.generate_slices(splitter.slabels, splitter.ngroups)

    # TODO: reset index
    sdata = splitter._get_sorted_data()

    # TODO: avoid costly make_block call, and hard-coded BlockManager init actions.
    #       neither of these things is necessary when subsetting rows.
    for start, end in zip(starts, ends):
        yield splitter._chop(sdata, slice(start, end))



@singledispatch2(pd.DataFrame)
def nest(__data, *args, key = "data"):
    """Nest columns within a DataFrame.
    

    Parameters
    ----------
    __data:
        A DataFrame.
    *args:
        The names of columns to be nested. May use any syntax used by the
        `select` function.
    key:
        The name of the column that will hold the nested columns.

    Examples
    --------

    >>> from siuba import _, nest
    >>> from siuba.data import cars
    >>> nested_cars = cars >> nest(-_.cyl)

    Note that pandas with nested DataFrames looks okay in juypter notebooks,
    but has a weird representation in the IPython console, so the example below
    shows that each entry in the data column is a DataFrame.

    >>> nested_cars.shape
    (3, 2)

    >>> type(nested_cars.data[0])
    <class 'pandas.core.frame.DataFrame'>
        
    """
    # TODO: copied from select function
    var_list = var_create(*args)
    od = var_select(__data.columns, *var_list)

    # unselected columns are treated similar to using groupby
    grp_keys = list(k for k in __data.columns if k not in set(od))
    nest_keys = list(od)

    # split into sub DataFrames, with only nest_keys as columns
    g_df = __data.groupby(grp_keys)
    splitter = g_df.grouper._get_splitter(g_df.obj[nest_keys])

    # TODO: iterating over splitter now only produces 1 item (the dataframe)
    # check backwards compat
    def _extract_subdf_pandas_1_3(entry):
        # in pandas < 1.3, splitter.__iter__ returns tuple entries (ii, df)
        if isinstance(entry, tuple):
            return entry[1]

        # in pandas 1.3, each entry is just the dataframe
        return entry

    result_index = g_df.grouper.result_index
    nested_dfs = [_extract_subdf_pandas_1_3(x) for x in splitter]

    out = pd.DataFrame({key: nested_dfs}, index = result_index).reset_index()

    return out

@nest.register(DataFrameGroupBy)
def _nest(__data, *args, key = "data"):
    from siuba.dply.tidyselect import VarAnd

    grp_keys = [x.name for x in __data.grouper.groupings]
    if None in grp_keys:
        raise NotImplementedError("All groupby variables must be named when using nest")

    sel_vars = var_create(*grp_keys)
    return nest(__data.obj, -VarAnd(sel_vars), *args, key = key)




# Unnest ======================================================================

@singledispatch2(pd.DataFrame)
def unnest(__data, key = "data"):
    """Unnest a column holding nested data (e.g. Series of lists or DataFrames).
    
    Parameters
    ----------
    ___data:
        A DataFrame.
    key:
        The name of the column to be unnested.

    Examples
    --------

    >>> import pandas as pd
    >>> df = pd.DataFrame({'id': [1,2], 'data': [['a', 'b'], ['c']]})
    >>> df >> unnest()
       id data
    0   1    a
    1   1    b
    2   2    c
        
    """
    # TODO: currently only takes key, not expressions
    nrows_nested = __data[key].apply(len, convert_dtype = True)
    indx_nested = nrows_nested.index.repeat(nrows_nested)

    grp_keys = list(__data.columns[__data.columns != key])

    # flatten nested data
    data_entries = map(_convert_nested_entry, __data[key])
    long_data = pd.concat(data_entries, ignore_index = True)
    long_data.name = key

    # may be a better approach using a multi-index
    long_grp = __data.loc[indx_nested, grp_keys].reset_index(drop = True)
    
    return long_grp.join(long_data)

def _convert_nested_entry(x):
    if isinstance(x, (tuple, list)):
        return pd.Series(x)
    
    return x


# Joins =======================================================================
from collections.abc import Mapping
from functools import partial
from pandas.core.reshape.merge import _MergeOperation


# TODO: will need to use multiple dispatch
@singledispatch2((pd.DataFrame, DataFrameGroupBy))
@_bounce_groupby
def join(left, right, on = None, how = None, *args, by = None, **kwargs):
    """Join two tables together, by matching on specified columns.

    The functions inner_join, left_join, right_join, and full_join are provided
    as wrappers around join, and are used in the examples.

    Parameters
    ----------
    left :
        The left-hand table.
    right :
        The right-hand table.
    on :
        How to match them. Note that the keyword "by" can also be used for this
        parameter, in order to support compatibility with dplyr.
    how :
        The type of join to perform (inner, full, left, right).
    *args:
        Additional postition arguments. Currently not supported.
    **kwargs:
        Additional keyword arguments. Currently not supported.


    Returns
    -------
    pd.DataFrame

    Examples
    --------

    >>> from siuba import _, inner_join, left_join, full_join, right_join
    >>> from siuba.data import band_members, band_instruments, band_instruments2
    >>> band_members
       name     band
    0  Mick   Stones
    1  John  Beatles
    2  Paul  Beatles

    >>> band_instruments
        name   plays
    0   John  guitar
    1   Paul    bass
    2  Keith  guitar

    Notice that above, only John and Paul have entries for band instruments.
    This means that they will be the only two rows in the inner_join result:

    >>> band_members >> inner_join(_, band_instruments)
       name     band   plays
    0  John  Beatles  guitar
    1  Paul  Beatles    bass

    A left join ensures all original rows of the left hand data are included.

    >>> band_members >> left_join(_, band_instruments)
       name     band   plays
    0  Mick   Stones     NaN
    1  John  Beatles  guitar
    2  Paul  Beatles    bass

    A full join is similar, but ensures all rows of both data are included.

    >>> band_members >> full_join(_, band_instruments)
        name     band   plays
    0   Mick   Stones     NaN
    1   John  Beatles  guitar
    2   Paul  Beatles    bass
    3  Keith      NaN  guitar

    You can explicilty specify columns to join on using the "by" argument:

    >>> band_members >> inner_join(_, band_instruments, by = "name")
       n...

    Use a dictionary for the by argument, to match up columns with different names:

    >>> band_members >> full_join(_, band_instruments2, {"name": "artist"})
       n...
 
    Joins create a new row for each pair of matches. For example, the value 1
    is in two rows on the left, and 2 rows on the right so 4 rows will be created.

    >>> df1 = pd.DataFrame({"x": [1, 1, 3]})
    >>> df2 = pd.DataFrame({"x": [1, 1, 2], "y": ["first", "second", "third"]})
    >>> df1 >> left_join(_, df2)
       x       y
    0  1   first
    1  1  second
    2  1   first
    3  1  second
    4  3     NaN

    Missing values count as matches to eachother by default:


    >>> df3 = pd.DataFrame({"x": [1, None], "y": 2})
    >>> df4 = pd.DataFrame({"x": [1, None], "z": 3})
    >>> left_join(df3, df4)
         x  y  z
    0  1.0  2  3
    1  NaN  2  3

    """

    if isinstance(right, DataFrameGroupBy):
        right = right.obj
    if not isinstance(right, DataFrame):
        raise Exception("right hand table must be a DataFrame")
    if how is None:
        raise Exception("Must specify how argument")

    if len(args) or len(kwargs):
        raise NotImplementedError("extra arguments to pandas join not currently supported")

    if on is None and by is not None:
        on = by

    # pandas uses outer, but dplyr uses term full
    if how == "full":
        how = "outer"

    if isinstance(on, Mapping):
        left_on, right_on = zip(*on.items())
        return left.merge(right, how = how, left_on = left_on, right_on = right_on)

    return left.merge(right, how = how, on = on)


@join.register(object)
def _join(left, right, on = None, how = None):
    raise Exception("Unsupported type %s" %type(left))


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
@_bounce_groupby
def semi_join(left, right = None, on = None, *args, by = None):
    """Return the left table with every row that would be kept in an inner join.

    Parameters
    ----------
    left :
        The left-hand table.
    right :
        The right-hand table.
    on :
        How to match them. By default it uses matches all columns with the same
        name across the two tables.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, semi_join, anti_join

    >>> df1 = pd.DataFrame({"id": [1, 2, 3], "x": ["a", "b", "c"]})
    >>> df2 = pd.DataFrame({"id": [2, 3, 3], "y": ["l", "m", "n"]})

    >>> df1 >> semi_join(_, df2)
       id  x
    1   2  b
    2   3  c

    >>> df1 >> anti_join(_, df2)
       id  x
    0   1  a

    Generally, it's a good idea to explicitly specify the on argument.

    >>> df1 >> anti_join(_, df2, on="id")
       id  x
    0   1  a
    """

    if on is None and by is not None:
        on = by

    if isinstance(on, Mapping):
        # coerce colnames to list, to avoid indexing with tuples
        on_cols, right_on = map(list, zip(*on.items()))
        right = right[right_on].rename(dict(zip(right_on, on_cols)))
    elif on is None:
        warnings.warn(
            "No on column passed to join. "
            "Inferring join columns instead using shared column names."
        )

        on_cols = list(set(left.columns).intersection(set(right.columns)))
        if not len(on_cols):
            raise Exception("No join column specified, and no shared column names")

        warnings.warn("Detected shared columns: %s" % on_cols)
    elif isinstance(on, str):
        on_cols = [on]
    else:
        on_cols = on

    # get our semi join on ----
    if len(on_cols) == 1:
        col_name = on_cols[0]
        indx = left[col_name].isin(right[col_name])
        return left.loc[indx]

    # Not a super efficient approach. Effectively, an inner join with what would
    # be duplicate rows removed.
    merger = _MergeOperation(left, right, left_on = on_cols, right_on = on_cols)
    _, l_indx, _ = merger._get_join_info()


    range_indx = pd.RangeIndex(len(left))
    return left.loc[range_indx.isin(l_indx)]


@singledispatch2((pd.DataFrame, DataFrameGroupBy))
@_bounce_groupby
def anti_join(left, right = None, on = None, *args, by = None):
    """Return the left table with every row that would *not* be kept in an inner join.

    Parameters
    ----------
    left :
        The left-hand table.
    right :
        The right-hand table.
    on :
        How to match them. By default it uses matches all columns with the same
        name across the two tables.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, semi_join, anti_join

    >>> df1 = pd.DataFrame({"id": [1, 2, 3], "x": ["a", "b", "c"]})
    >>> df2 = pd.DataFrame({"id": [2, 3, 3], "y": ["l", "m", "n"]})

    >>> df1 >> semi_join(_, df2)
       id  x
    1   2  b
    2   3  c

    >>> df1 >> anti_join(_, df2)
       id  x
    0   1  a

    Generally, it's a good idea to explicitly specify the on argument.

    >>> df1 >> anti_join(_, df2, on="id")
       id  x
    0   1  a
    """

    if on is None and by is not None:
        on = by

    # copied from semi_join
    if isinstance(on, Mapping):
        left_on, right_on = zip(*on.items())
    else: 
        left_on = right_on = on

    if isinstance(right, DataFrameGroupBy):
        right = right.obj

    # manually perform merge, up to getting pieces need for indexing
    merger = _MergeOperation(left, right, left_on = left_on, right_on = right_on)
    _, l_indx, _ = merger._get_join_info()
    
    # use the left table's indexer to exclude those rows
    range_indx = pd.RangeIndex(len(left))
    return left.iloc[range_indx.difference(l_indx),:]

left_join = partial(join, how = "left")
right_join = partial(join, how = "right")
full_join = partial(join, how = "full")
inner_join = partial(join, how = "inner")


# Head ========================================================================

@singledispatch2(pd.DataFrame)
def head(__data, n = 5):
    """Return the first n rows of the data.

    Parameters
    ----------
    __data:
        a DataFrame.
    n:
        The number of rows of data to keep.

    Examples
    --------

    >>> from siuba import head
    >>> from siuba.data import cars

    >>> cars >> head(2)
       cyl   mpg   hp
    0    6  21.0  110
    1    6  21.0  110
    """

    return __data.head(n)


@head.register(DataFrameGroupBy)
def _head_gdf(__data, n = 5):
    groupings = __data.grouper.groupings
    group_cols = [ping.name for ping in groupings]

    df_subset = __data.obj.head(n)
    return df_subset.groupby(group_cols)


# Top N =======================================================================

# TODO: should dispatch to filter, no need to specify pd.DataFrame?
@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def top_n(__data, n, wt = None):
    """Filter to keep the top or bottom entries in each group.

    Parameters
    ----------
    ___data:
        A DataFrame.
    n:
        The number of rows to keep in each group.
    wt:
        A column or expression that determines ordering (defaults to last column in data).

    Examples
    --------
    >>> from siuba import _, top_n
    >>> df = pd.DataFrame({'x': [3, 1, 2, 4], 'y': [1, 1, 0, 0]})
    >>> top_n(df, 2, _.x)
       x  y
    0  3  1
    3  4  0

    >>> top_n(df, -2, _.x)
       x  y
    1  1  1
    2  2  0

    >>> top_n(df, 2, _.x*_.y)
       x  y
    0  3  1
    1  1  1

    """
    # NOTE: using min_rank, since it can return a lazy expr for min_rank(ing)
    #       but I would rather not have it imported in verbs. will be more 
    #       reasonable if each verb were its own file? need abstract verb / vector module.
    #       vector imports experimental right now, so we need to invert deps
    # TODO: 
    #   * what if wt is a string? should remove all str -> expr in verbs like group_by etc..
    #   * verbs like filter allow lambdas, but this func breaks with that   
    from .vector import min_rank
    if wt is None:
        sym_wt = getattr(Symbolic(MetaArg("_")), __data.columns[-1])
    elif isinstance(wt, Call):
        sym_wt = Symbolic(wt)
    else:
        raise TypeError("wt must be a symbolic expression, eg. _.some_col")

    if n > 0:
        return filter(__data, min_rank(-sym_wt) <= n)
    else:
        return filter(__data, min_rank(sym_wt) <= abs(n))


# Gather ======================================================================

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def gather(__data, key = "key", value = "value", *args, drop_na = False, convert = False):
    """Reshape table by gathering it in to long format.

    Parameters
    ----------
    __data:
        The input data.
    key:
        Name of the key (or measure) column, which holds the names of the columns
        that were turned into rows.
    value:
        Name of the value column, which holds the values from the columns that
        were turned into rows.
    *args:
        A selection of columns. If unspecified, all columns are selected. Any
        arguments you could pass to the select() verb are allowed.
    drop_na: bool
        Whether to remove any rows where the value column is NA.
        

    Examples
    --------

    >>> import pandas as pd
    >>> from siuba import _, gather

    >>> df = pd.DataFrame({"id": ["a", "b"], "x": [1, 2], "y": [3, None]})

    The code below gathers in all columns, except id:

    >>> gather(df, "key", "value", -_.id)
      id key  value
    0  a   x    1.0
    1  b   x    2.0
    2  a   y    3.0
    3  b   y    NaN

    >>> gather(df, "measure", "result", _.x, _.y, drop_na=True)
      id measure  result
    0  a       x     1.0
    1  b       x     2.0
    2  a       y     3.0

    """
    # TODO: implement var selection over *args
    if convert:
        raise NotImplementedError("convert not yet implemented")

    # TODO: copied from nest and select
    var_list = var_create(*(args or __data.columns))
    od = var_select(__data.columns, *var_list)

    if not od:
        return __data

    id_vars = [col for col in __data.columns if col not in od]
    long = pd.melt(__data, id_vars, list(od), key, value)

    if drop_na:
        return long[~long[value].isna()].reset_index(drop = True)

    return long


@gather.register(DataFrameGroupBy)
def _gather(__data, key = "key", value = "value", *args, **kwargs):
    group_cols = [ping.name for ping in __data.grouper.groupings]

    res = gather(__data.obj, key, value, *args, **kwargs)

    # regroup on any grouping vars we did not gather ----
    candidates = set(res.columns) - {key, value}
    regroup_cols = [name for name in group_cols if name in candidates]

    if res is __data.obj:
        # special case where nothing happened
        return __data
    elif regroup_cols:
        return res.groupby(regroup_cols)

    return res


# Spread ======================================================================

def _get_single_var_select(columns, x):
    od = var_select(columns, *var_create(x))

    if len(od) != 1:
        raise ValueError("Expected single variable, received: %s" %list(od))

    return next(iter(od))

@singledispatch2(pd.DataFrame)
def spread(__data, key, value, fill = None, reset_index = True):
    """Reshape table by spreading it out to wide format.

    Parameters
    ----------
    __data:
        The input data.
    key:
        Column whose values will be used as new column names.
    value:
        Column whose values will fill the new column entries.
    fill:
        Value to set for any missing values. By default keeps them as missing values.


    Examples
    --------
    >>> import pandas as pd                                                
    >>> from siuba import _, gather                                        

    >>> df = pd.DataFrame({"id": ["a", "b"], "x": [1, 2], "y": [3, None]}) 

    >>> long = gather(df, "key", "value", -_.id, drop_na=True)
    >>> long
      id key  value
    0  a   x    1.0
    1  b   x    2.0
    2  a   y    3.0

    >>> spread(long, "key", "value")
      id    x    y
    0  a  1.0  3.0
    1  b  2.0  NaN

    """
    key_col = _get_single_var_select(__data.columns, key)
    val_col = _get_single_var_select(__data.columns, value)

    id_cols = [col for col in __data.columns if col not in (key_col, val_col)]
    wide = __data.set_index(id_cols + [key_col]).unstack(level = -1)
    
    if fill is not None:
        wide.fillna(fill, inplace = True)
    
    # remove multi-index from both rows and cols
    wide.columns = wide.columns.droplevel().rename(None)
    if reset_index:
        wide.reset_index(inplace = True)
    
    return wide


@spread.register(DataFrameGroupBy)
def _spread_gdf(__data, *args, **kwargs):

    groupings = __data.grouper.groupings

    df = __data.obj

    f_spread = spread.registry[pd.DataFrame]
    out = f_spread(df, *args, **kwargs)

    # regroup, using group names ----
    group_names = [x.name for x in groupings]
    if any([name is None for name in group_names]):
        raise ValueError("spread can only work on grouped DataFrame if all groupings "
                         "have names. Groups are: %s" %group_names)

    return out.groupby(group_names)

# Expand/Complete ====================================================================
from pandas.core.reshape.util import cartesian_product


def _unique_name(prefix: str, names: "set[str]"):
    names = set(names)

    ii = 0
    while prefix in names:
        prefix = prefix + str(ii)
        
        ii += 1

    return prefix


def _expand_column(x):
    from pandas.api.types import is_categorical_dtype

    if is_categorical_dtype(x):
        if x.isna().any():
            return [*x.cat.categories, None]

        return x.cat.categories

    return x.unique()



@singledispatch2(pd.DataFrame)
def expand(__data, *args, fill = None):
    """Return table with unique crossings of specified columns.

    Parameters
    ----------
    __data:
        The input data.
    *args:
        Column names to cross and de-duplicate.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, expand, count, anti_join, right_join

    >>> df = pd.DataFrame({"x": [1, 2, 2], "y": ["a", "a", "b"], "z": 1})
    >>> df
       x  y  z
    0  1  a  1
    1  2  a  1
    2  2  b  1

    >>> combos = df >> expand(_.x, _.y)
    >>> combos
       x  y
    0  1  a
    1  1  b
    2  2  a
    3  2  b

    >>> df >> right_join(_, combos)
       x  y    z
    0  1  a  1.0
    1  1  b  NaN
    2  2  a  1.0
    3  2  b  1.0

    >>> combos >> anti_join(_, df)
       x  y
    1  1  b

    Note that expand will also cross missing values: 

    >>> df2 = pd.DataFrame({"x": [1, None], "y": [3, 4]})
    >>> expand(df2, _.x, _.y)
         x  y
    0  1.0  3
    1  1.0  4
    2  NaN  3
    3  NaN  4

    It will also cross all levels of a categorical (even those not in the data):

    >>> df3 = pd.DataFrame({"x": pd.Categorical(["a"], ["a", "b"])})
    >>> expand(df3, _.x)
       x
    0  a
    1  b

    """


    var_names = list(map(simple_varname, args))
    cols = [_expand_column(__data.loc[:, name]) for name in var_names]

    if fill is not None:
        raise NotImplementedError()

    return pd.MultiIndex.from_product(cols, names=var_names).to_frame(index=False)


@singledispatch2(pd.DataFrame)
def complete(__data, *args, fill = None, explicit=True):
    """Add rows to fill in missing combinations in the data.

    This is a wrapper around expand(), right_join(), along with filling NAs.

    Parameters
    ----------
    __data:
        The input data.
    *args:
        Columns to cross and expand.
    fill:
        A dictionary specifying what to use for missing values in each column.
        If a column is not specified, missing values are left as is.
    explicit:
        Should both NAs created by the complete and pre-existing NAs be filled
        by the fill argument? Defaults to True (filling both). When set to False,
        it will only fill newly created NAs.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import _, expand, count, anti_join, right_join

    >>> df = pd.DataFrame({"x": [1, 2, 2], "y": ["a", "a", "b"], "z": [8, 9, None]})
    >>> df
       x  y    z
    0  1  a  8.0
    1  2  a  9.0
    2  2  b  NaN

    >>> df >> complete(_.x, _.y)
       x  y    z
    0  1  a  8.0
    1  1  b  NaN
    2  2  a  9.0
    3  2  b  NaN

    Use the fill argument to replace missing values:

    >>> df >> complete(_.x, _.y, fill={"z": 999})
       x  y      z
    0  1  a    8.0
    1  1  b  999.0
    2  2  a    9.0
    3  2  b  999.0

    A common use of complete is to make zero counts explicit (e.g. for charting):

    >>> df >> count(_.x, _.y) >> complete(_.x, _.y, fill={"n": 0})
       x  y    n
    0  1  a  1.0
    1  1  b  0.0
    2  2  a  1.0
    3  2  b  1.0
    
    Use explicit=False to only fill the NaNs introduced by complete (implicit missing),
    and not those already in the original data (explicit missing):

    >>> df >> complete(_.x, _.y, fill={"z": 999}, explicit=False)
       x  y      z
    0  1  a    8.0
    1  1  b  999.0
    2  2  a    9.0
    3  2  b    NaN
    
    """

    if explicit:
        indicator = False
    else:
        indicator = _unique_name("__merge_indicator", {*__data.columns})
    

    expanded = expand(__data, *args)

    # TODO: should we attempt to coerce cols back to original types?
    #       e.g. NAs will turn int -> float
    on_cols = list(expanded.columns)
    df = expanded.merge(__data, how = "outer", on = on_cols, indicator = indicator)
    
    if fill is not None:
        if explicit:
            for col_name, val in fill.items():
                df[col_name].fillna(val, inplace = True)
        else:
            fill_cols = list(fill)
            indx = df[indicator] == "left_only"
            df.loc[indx, fill_cols] = df.loc[indx, fill_cols].fillna(fill)

    if indicator:
        return df.drop(columns=indicator)

    return df
    

# Separate/Unit/Extract ============================================================

@singledispatch2(pd.DataFrame)
def separate(__data, col, into, sep = r"[^a-zA-Z0-9]",
             remove = True, convert = False,
             extra = "warn", fill = "warn"
            ):
    """Split col into len(into) piece. Return DataFrame with a column added for each piece.

    Parameters
    ----------
    __data:
        a DataFrame.
    col:
        name of column to split (either string, or siu expression).
    into:
        names of resulting columns holding each entry in split.
    sep:
        regular expression used to split col. Passed to col.str.split method.
    remove:
        whether to remove col from the returned DataFrame.
    convert:
        whether to attempt to convert the split columns to numerics.
    extra:
        what to do when more splits than into names.  One of ("warn", "drop" or "merge").
        "warn" produces a warning; "drop" and "merge" currently not implemented.
    fill:
        what to do when fewer splits than into names. Currently not implemented.

    Examples
    --------
    >>> import pandas as pd
    >>> from siuba import separate

    >>> df = pd.DataFrame({"label": ["S1-1", "S2-2"]})

    Split into two columns:

    >>> separate(df, "label", into = ["season", "episode"])
      season episode
    0     S1       1
    1     S2       2

    Split, and try to convert columns to numerics:
    
    >>> separate(df, "label", into = ["season", "episode"], convert = True)
      season  episode
    0     S1        1
    1     S2        2

    """

    n_into = len(into)
    col_name = simple_varname(col)
    
    # splitting column ----
    all_splits = __data[col_name].str.split(sep, expand = True)
    n_split_cols = len(all_splits.columns)
    
    # handling too many or too few splits ----
    if  n_split_cols < n_into:
        # too few columns
        raise ValueError("Expected %s split cols, found %s" %(n_into, n_split_cols))
    elif n_split_cols > n_into:
        # Extra argument controls how we deal with too many splits
        if extra == "warn":
            df_extra_cols = all_splits.iloc[:, n_into].reset_index(drop=True)
            bad_rows = df_extra_cols.dropna(how="all")
            n_extra = bad_rows.shape[0]

            warnings.warn(
                f"Expected {n_into} pieces."
                f"Additional pieces discarded in {n_extra} rows."
                f"Row numbers: {bad_rows.index.values}",
                UserWarning
            )
        elif extra == "drop":
            pass
        elif extra == "merge":
            raise NotImplementedError("TODO: separate extra = 'merge'")
        else:
            raise ValueError("Invalid extra argument: %s" %extra)

    # create new columns in data ----
    out = __data.copy()

    for ii, name in enumerate(into):
        out[name] = all_splits.iloc[:, ii]
    
    #out = pd.concat([__data, keep_splits], axis = 1)

    # attempt to convert columns to numeric ----
    if convert:
        # TODO: better strategy here? 
        for k in into:
            try:
                out[k] = pd.to_numeric(out[k])
            except ValueError:
                pass

    if remove and col_name not in into:
        return out.drop(columns = col_name)

    return out


@separate.register(DataFrameGroupBy)
def _separate_gdf(__data, *args, **kwargs):

    groupings = __data.grouper.groupings

    df = __data.obj

    f_separate = separate.registry[pd.DataFrame]
    out = f_separate(df, *args, **kwargs)

    return out.groupby(groupings)


def _coerce_to_str(arr):
    """Return either original series, or ser.astype(str)"""
    if pd.api.types.is_string_dtype(arr):
        return arr

    return arr.astype(str)


# unite ----

from functools import reduce

@singledispatch2(pd.DataFrame)
def unite(__data, col, *args, sep = "_", remove = True):
    """Combine multiple columns into a single column. Return DataFrame that column included.

    Parameters
    ----------
    __data:
        a DataFrame
    col:
        name of the to-be-created column (string).
    *args:
        names of each column to combine.
    sep:
        separator joining each column being combined.
    remove:
        whether to remove the combined columns from the returned DataFrame.

    """
    unite_col_names = list(map(simple_varname, args))
    out_col_name = simple_varname(col)

    # validations ----
    if None in unite_col_names:
        raise ValueError("*args must be string, or simple column name, e.g. _.col_name")

    missing_cols = set(unite_col_names) - set(__data.columns)
    if missing_cols:
        raise ValueError("columns %s not in DataFrame.columns" %missing_cols)


    unite_cols = [_coerce_to_str(__data[col_name]) for col_name in unite_col_names]

    if out_col_name in __data:
        raise ValueError("col argument %s already a column in data" % out_col_name)

    # perform unite ----
    # TODO: this is probably not very efficient. Maybe try with transform or apply?
    res = reduce(lambda x,y: x + sep + y, unite_cols)

    out_df = __data.copy()
    out_df[out_col_name] = res

    if remove:
        return out_df.drop(columns = unite_col_names)

    return out_df

@unite.register(DataFrameGroupBy)
def _unite_gdf(__data, *args, **kwargs):
    # TODO: consolidate these trivial group by dispatched funcs

    groupings = __data.grouper.groupings

    df = __data.obj

    f_unite = unite.registry[pd.DataFrame]
    out = f_unite(df, *args, **kwargs)

    return out.groupby(groupings)


# extract ----

@singledispatch2(pd.DataFrame)
def extract(
        __data, col, into, regex = r"(\w+)",
        remove = True, convert = False,
        flags = 0
        ):
    """Pull out len(into) fields from character strings. 

    Returns a DataFrame with a column added for each piece.

    Parameters
    ----------
    __data:
        a DataFrame
    col:
        name of column to split (either string, or siu expression).
    into:
        names of resulting columns holding each entry in pulled out fields.
    regex:
        regular expression used to extract field. Passed to col.str.extract method.
    remove:
        whether to remove col from the returned DataFrame.
    convert:
        whether to attempt to convert the split columns to numerics.
    flags:
        flags from the re module, passed to col.str.extract.

    """

    col_name = simple_varname(col)
    n_into = len(into)

    all_splits = __data[col_name].str.extract(regex, flags)
    n_split_cols = len(all_splits.columns)

    if n_split_cols != n_into:
        raise ValueError("Split into %s pieces, but expected %s" % (n_split_cols, n_into))

    # attempt to convert columns to numeric ----
    if convert:
        # TODO: better strategy here? 
        for k in all_splits:
            try:
                all_splits[k] = pd.to_numeric(all_splits[k])
            except ValueError:
                pass

    out = __data.copy()
    for ii, name in enumerate(into):
        out[name] = all_splits.iloc[:, ii]
    
    if remove:
        return out.drop(columns = col_name)

    return out

@extract.register(DataFrameGroupBy)
def _extract_gdf(__data, *args, **kwargs):
    # TODO: consolidate these trivial group by dispatched funcs

    groupings = __data.grouper.groupings

    df = __data.obj

    f_extract = extract.registry[pd.DataFrame]
    out = f_extract(df, *args, **kwargs)

    return out.groupby(groupings)


# tbl ----

from siuba.siu._databackend import SqlaEngine, PlDataFrame, PdDataFrame

@singledispatch2((pd.DataFrame, DataFrameGroupBy))
def tbl(src, *args, **kwargs):
    """Create a table from a data source.

    Parameters
    ----------
    src:
        A pandas DataFrame, SQLAlchemy Engine, or other registered object.
    *args, **kwargs:
        Additional arguments passed to the individual implementations.

    Examples
    --------
    >>> from siuba.data import cars

    A pandas DataFrame is already a table of data, so trivially returns itself.

    >>> tbl(cars) is cars
    True

    tbl() is useful for quickly connecting to a SQL database table.

    >>> from sqlalchemy import create_engine
    >>> from siuba import count, show_query, collect

    >>> engine = create_engine("sqlite:///:memory:")
    >>> _rows = cars.to_sql("cars", engine, index=False)

    >>> tbl_sql_cars = tbl(engine, "cars")
    >>> tbl_sql_cars >> count()
    # Source: lazy query
    # DB Conn: Engine(sqlite:///:memory:)
    # Preview:
        n
    0  32
    # .. may have more rows

    When using duckdb, pass a DataFrame as the third argument to operate directly on it:

    >>> engine2 = create_engine("duckdb:///:memory:")
    >>> tbl_cars_duck = tbl(engine, "cars", cars.head(2)) 
    >>> tbl_cars_duck >> count() >> collect()
        n
    0  32

    You can analyze a mock table

    >>> from sqlalchemy import create_mock_engine
    >>> from siuba import _

    >>> mock_engine = create_mock_engine("postgresql:///", lambda *args, **kwargs: None)
    >>> tbl_mock = tbl(mock_engine, "some_table", columns = ["a", "b", "c"])

    >>> q = tbl_mock >> count(_.a) >> show_query()    # doctest: +NORMALIZE_WHITESPACE
    SELECT some_table_1.a, count(*) AS n
    FROM some_table AS some_table_1 GROUP BY some_table_1.a ORDER BY n DESC
    """

    return src

@tbl.register
def _tbl_sqla(src: SqlaEngine, table_name, columns=None):
    from siuba.sql import LazyTbl

    # TODO: once we subclass LazyTbl per dialect (e.g. duckdb), we can move out
    # this dialect specific logic.
    if src.dialect.name == "duckdb" and isinstance(columns, (PdDataFrame, PlDataFrame)):
        with src.begin() as conn:
            conn.exec_driver_sql("register", (table_name, columns))
        
        return LazyTbl(src, table_name)
    
    return LazyTbl(src, table_name, columns=columns)


@tbl.register(object)
def _tbl(__data, *args, **kwargs):
    # sqlalchemy v2 does not have MockConnection inherit from anything
    # even though it is a mock :/.
    try:
        from sqlalchemy.engine.mock import MockConnection

        if isinstance(__data, MockConnection):
            return tbl.dispatch(SqlaEngine)(__data, *args, **kwargs)
    except ImportError:
        pass

    raise NotImplementedError(
        f"Unsupported type {type(__data)}. "
        "Note that tbl currently can be used at the start of a pipe, but not as "
        "a step in the pipe."
    )

# Install Siu =================================================================

install_pd_siu()



# sqlvariant, allow defining 3 namespaces to override defaults
from ..translate import (
        SqlColumn, SqlColumnAgg, extend_base, win_agg,
        SqlTranslator, sql_not_impl, win_absent
        )

import sqlalchemy.sql.sqltypes as sa_types

from sqlalchemy import sql
from sqlalchemy.sql import func as fn

# Custom dispatching in call trees ============================================

class MysqlColumn(SqlColumn): pass
class MysqlColumnAgg(SqlColumnAgg, MysqlColumn): pass

def sql_str_strip(left = True, right = True):
    def f(_, col):
        # see https://stackoverflow.com/a/6858168/1144523
        lstrip = "^[[:space:]]+" if left else ""
        rstrip = "[[:space:]]+$" if right else ""
    
        or_op = "|" if lstrip and rstrip else ""
        regex = "(" + lstrip + or_op + rstrip + ")"
    
        return fn.regexp_replace(col, regex, "")

    return f

# Date functions --------------------------------------------------------------

from . import _dt_generics as _dt

def sql_func_extract_dow_monday(_, col):
    # MYSQL: sunday starts, equals 1 (an int)
    # pandas: monday starts, equals 0 (also an int)

    raw_dow = fn.dayofweek(col)

    # monday is 2 in MYSQL, so use monday + 5 % 7
    return (raw_dow + 5) % 7


@_dt.sql_is_first_day_of.register(MysqlColumn)
def sql_is_first_day_of(_, col, period):
    src_per = fn.extract(period, col)
    incr_per = fn.extract(period, fn.date_sub(col, sql.text("INTERVAL 1 DAY")))

    return src_per != incr_per

@_dt.sql_is_last_day_of.register(MysqlColumn)
def sql_is_last_day_of(_, col, period):
    src_per = fn.extract(period, col)
    incr_per = fn.extract(period, fn.date_add(col, sql.text("INTERVAL 1 DAY")))

    return src_per != incr_per
    

def sql_func_truediv(_, x, y):
    return sql.cast(x, sa_types.Numeric()) / y

def sql_func_floordiv(_, x, y):
    return x.op("DIV")(y)

def sql_func_between(_, col, left, right, inclusive=True):
    if not inclusive:
        raise NotImplementedError("between must be inclusive")

    # TODO: should figure out how sqlalchemy prefers to set types, rather
    # than setting manually on this expression
    expr = col.between(left, right)
    expr.type = sa_types.Boolean()
    return expr

scalar = extend_base(
        MysqlColumn,

        # copied from postgres. MYSQL does true division over ints by default,
        # but it does not produce double precision.
        div = sql_func_truediv,
        divide = sql_func_truediv,
        rdiv = lambda codata, x,y: sql_func_truediv(codata, y, x),

        __truediv__ = sql_func_truediv,
        truediv = sql_func_truediv,
        __rtruediv__ = lambda codata, x, y: sql_func_truediv(codata, y, x),

        __floordiv__ = sql_func_floordiv,
        __rfloordiv__ = lambda codata, x, y: sql_func_floordiv(codata, y, x),

        between = sql_func_between,

        **{
          "str.lstrip": sql_str_strip(right = False),
          "str.rstrip": sql_str_strip(left = False),
          "str.strip": sql_str_strip(),
          "str.title": sql_not_impl()                       # see https://stackoverflow.com/q/12364086/1144523
        },
        **{
          "dt.dayofyear": lambda _, col: fn.dayofyear(col),
          "dt.dayofweek": sql_func_extract_dow_monday,
          "dt.days_in_month": lambda _, col: fn.dayofmonth(fn.last_day(col)),
          "dt.daysinmonth": lambda _, col: fn.dayofmonth(fn.last_day(col)),
          "dt.is_month_end": lambda _, col: col == fn.last_day(col),
          "dt.is_month_start": lambda _, col: fn.dayofmonth(col) == 1,
          "dt.week": lambda _, col: fn.week(col, 1),
          "dt.weekday": sql_func_extract_dow_monday,
          "dt.weekofyear": lambda _, col: fn.week(col, 1),
        }
        )

aggregate = extend_base(
        MysqlColumnAgg,

        quantile = win_absent("percentile_cont"),
        )

window = extend_base(
        MysqlColumn,

        # TODO: analytic percentile_cont is supported in mariadb
        quantile = win_absent("percentile_cont"),
        #sd = win_agg("stddev")
        )

funcs = dict(scalar = scalar, aggregate = aggregate, window = window)

translator = SqlTranslator.from_mappings(
        MysqlColumn, MysqlColumnAgg
        )




import pandas as pd
import pytest

from pandas.testing import assert_frame_equal
from pandas.core.groupby import DataFrameGroupBy
from siuba.siu import symbolic_dispatch, Symbolic, Fx
from siuba.dply import verbs
from siuba.dply.verbs import mutate, filter, summarize, group_by, collect, ungroup
from siuba.dply.across import across

from siuba.experimental.pivot.test_pivot import assert_equal_query2
from siuba.sql.translate import SqlColumn, SqlColumnAgg, sql_scalar, win_agg, sql_agg
from siuba.tests.helpers import assert_frame_sort_equal


# Helpers =====================================================================

_ = Symbolic()

# round function ----

@symbolic_dispatch(cls = pd.Series)
def f_round(x) -> pd.Series:
    return round(x)


f_round.register(SqlColumn, sql_scalar("round"))

# mean function ----

@symbolic_dispatch(cls = pd.Series)
def f_mean(x) -> pd.Series:
    return x.mean()

f_mean.register(SqlColumn, win_agg("avg"))
f_mean.register(SqlColumnAgg, sql_agg("avg"))


def assert_grouping_names(gdf, names):
    from siuba.sql import LazyTbl

    if isinstance(gdf, LazyTbl):
        grouping_names = list(gdf.group_by)
    else:
        assert isinstance(gdf, DataFrameGroupBy)
        groupings = gdf.grouper.groupings
        grouping_names = [g.name for g in groupings]

    assert len(grouping_names) == len(names)
    assert grouping_names == names


# Fixtures ====================================================================

@pytest.fixture
def df():
    return pd.DataFrame({
        "a_x": [1, 2],
        "a_y": [3, 4],
        "b_x": [5., 6.],    # note the floats
        "g": ["m", "n"]
    })


# Tests =======================================================================

TRANSFORMATION_FUNCS = [
    f_round,
    Fx.round(),
    f_round(Fx),
]


@pytest.mark.parametrize("func", TRANSFORMATION_FUNCS)
def test_across_func_transform(df, func):
    res = across(df, _[_.a_x, _.a_y], func)
    dst = pd.DataFrame({
        "a_x": df.a_x.round(),
        "a_y": df.a_y.round()
    })

    assert_equal_query2(res, dst)


def test_across_func_transform_lambda(df):
    res = across(df, _[_.a_x, _.a_y], lambda x: x.round())
    dst = pd.DataFrame({
        "a_x": df.a_x.round(),
        "a_y": df.a_y.round()
    })

    assert_frame_equal(res, dst)


@pytest.mark.parametrize("func", [
    (f_mean),
    (Fx.mean()),
    (f_mean(Fx)),
    (lambda x: x.mean()),
])
def test_across_func_aggregate(df, func):
    res = across(df, _[_.a_x, _.a_y], func)
    dst = pd.DataFrame({
        "a_x": [df.a_x.mean()],
        "a_y": [df.a_y.mean()]
    })

    assert_frame_equal(res, dst)


@pytest.mark.parametrize("func", [
    (lambda x: x % 2 > 1),
    (Fx % 2 > 1),
])
def test_across_func_bool(df, func):
    res = across(df, _[_.a_x, _.a_y], func)
    dst = pd.DataFrame({
        "a_x": df.a_x % 2 > 1,
        "a_y": df.a_y % 2 > 1
    })

    assert_frame_equal(res, dst)


@pytest.mark.parametrize("selection", [
    (_[0,1]),
    (_[0:"a_y"]),
    (lambda x: x.dtype == "int64"),
    #(where(Fx.dtype == "int64")),
    #(where(Fx.dtype != "float64") & ~_.g),
    (~_[_.b_x, _.g]),

])
def test_across_selection(df, selection):
    res = across(df, selection, lambda x: x + 1)
    dst = df[["a_x", "a_y"]] + 1

    assert_frame_equal(res, dst)


def test_across_selection_rename(df):
    res = across(df, _.zzz == _.a_x, lambda x: x + 1)
    assert res.columns.tolist() == ["zzz"]

    assert_frame_equal(res, (df[["a_x"]] + 1).rename(columns={"a_x": "zzz"}))


@pytest.mark.parametrize("func", TRANSFORMATION_FUNCS)
def test_across_in_mutate(backend, df, func):
    src = backend.load_df(df)
    res_explicit = mutate(src, across(_, _[_.a_x, _.a_y], f_round))
    res_implicit = mutate(src, across(_[_.a_x, _.a_y], f_round))

    dst = df.copy()
    dst["a_x"] = df.a_x.round()
    dst["a_y"] = df.a_y.round()

    sql_kwargs = {"check_dtype": False}
    assert_equal_query2(res_explicit, dst, sql_kwargs=sql_kwargs)
    assert_equal_query2(res_implicit, dst, sql_kwargs=sql_kwargs)


def test_across_in_mutate_grouped_equiv_ungrouped(backend, df):
    src = backend.load_df(df)
    g_src = group_by(src, "g")

    expr_across = across(_, _[_.a_x, _.a_y], f_round)
    g_res = mutate(g_src, expr_across)
    dst = mutate(src, expr_across)

    assert_grouping_names(g_res, ["g"])
    assert_equal_query2(ungroup(g_res), collect(dst))


def test_across_in_mutate_grouped_agg(backend):
    df = pd.DataFrame({"x": [1, 2, 3], "y": [7, 8, 9], "g": [1, 1, 2]})
    src = backend.load_df(df)
    g_src = group_by(src, "g")

    expr_across = across(_, _[_.x, _.y], f_mean)
    g_res = mutate(g_src, expr_across)
    dst = mutate(df.groupby("g"), expr_across)

    assert_grouping_names(g_res, ["g"])
    assert_equal_query2(ungroup(g_res), ungroup(dst))


def test_across_in_mutate_grouped_transform(backend):
    df = pd.DataFrame({"x": [1, 2, 3], "y": [7, 8, 9], "g": [1, 1, 2]})
    src = backend.load_df(df)
    g_src = group_by(src, "g")

    expr_across = across(_, _[_.x, _.y], Fx.rank())
    g_res = mutate(g_src, expr_across)
    dst = pd.DataFrame({"x": [1., 2, 1], "y": [1., 2, 1], "g": [1, 1, 2]})

    assert_grouping_names(g_res, ["g"])
    assert_equal_query2(ungroup(g_res), dst)

def test_across_in_summarize(backend, df):
    src = backend.load_df(df)
    res = summarize(src, across(_, _[_.a_x, _.a_y], f_mean))
    dst = pd.DataFrame({
        "a_x": [df.a_x.mean()],
        "a_y": [df.a_y.mean()]
    })

    assert_equal_query2(res, dst)


def test_across_in_summarize_equiv_ungrouped(backend):
    # note that summarize does not automatically regroup on any keys
    df = pd.DataFrame({
        "a_x": [1, 2],
        "a_y": [3, 4],
        "b_x": [5., 6.],
        "g": ["ZZ", "ZZ"]   # Note: all groups the same
    })
    src = backend.load_df(df)

    g_src = group_by(src, _.g)

    expr_across = across(_, _[_.a_x, _.a_y], f_mean)
    g_res = summarize(g_src, expr_across)
    dst = summarize(src, expr_across)

    
    collected = collect(g_res)
    assert collected.columns.tolist() == ["g", "a_x", "a_y"]
    assert collected["g"].tolist() == ["ZZ"]

    assert_frame_sort_equal(collected.drop(columns="g"), collect(dst))


def test_across_in_summarize_grouped(backend):
    df = pd.DataFrame({"x": [1, 2, 3], "y": [7, 8, 9], "g": [1, 1, 2]})
    src = backend.load_df(df)
    g_src = group_by(src, "g")

    expr_across = across(_, _[_.x, _.y], f_mean)
    g_res = summarize(g_src, expr_across)
    dst = pd.DataFrame({"g": [1, 2], "x": [1.5, 3], "y": [7.5, 9]})

    assert_equal_query2(ungroup(g_res), dst)


def test_across_in_filter(backend, df):
    src = backend.load_df(df)
    res = filter(src, across(_, _[_.a_x, _.a_y], Fx % 2 > 0))

    dst = df[(df[["a_x", "a_y"]] % 2 > 0).all(axis=1)]

    assert_equal_query2(res, dst)


def test_across_in_filter_equiv_ungrouped(backend, df):
    src = backend.load_df(df)

    expr_across = across(_, _[_.a_x, _.a_y], lambda x: x % 2 > 0)
    g_res = filter(group_by(df, _.g), expr_across)
    dst = filter(df, expr_across)

    assert_grouping_names(g_res, ["g"])
    assert_equal_query2(g_res.obj, dst)


def test_across_in_filter_grouped(backend):
    df = pd.DataFrame({"x": [1, 2, 3], "y": [7, 8, 9], "g": [1, 1, 2]})
    src = backend.load_df(df)
    g_src = group_by(src, "g")

    expr_across = across(_, _[_.x, _.y], Fx >= Fx.mean())
    g_res = filter(g_src, expr_across)
    dst = pd.DataFrame({"x": [2, 3], "y": [8, 9], "g": [1, 2]}, index=[1, 2])

    assert_equal_query2(ungroup(g_res), dst)


@pytest.mark.parametrize("f", [
    #(arrange),
    (verbs.count),
    (verbs.add_count),
    (verbs.distinct),
    (verbs.group_by),
    (verbs.transmute),
    
])
def test_across_in_verb(backend, df, f):
    src = backend.load_df(df)
    expr_across = across(_, _[_.a_x, _.a_y], Fx % 2 > 0)
    expr_manual = {"a_x": _.a_x % 2 > 0, "a_y": _.a_y % 2 > 0}

    res = f(src, expr_across)
    dst = f(df, **expr_manual)

    if isinstance(dst, DataFrameGroupBy):
        assert_grouping_names(res, ["a_x", "a_y"])

    assert_equal_query2(ungroup(res), ungroup(dst))


def test_across_in_arrange_unsupported(backend, df):
    src = backend.load_df(df)
    expr_across = across(_, _[_.a_x, _.a_y], Fx % 2 > 0)

    with pytest.raises(NotImplementedError) as exc_info:
        res = verbs.arrange(src, expr_across)

    assert "`arrange()` expression 0 of 1 returned" in exc_info.value.args[0]



def test_across_formula_and_underscore(df):
    res = across(df, _[_.a_x, _.a_y], f_round(Fx) / _.b_x)

    dst = pd.DataFrame({
        "a_x": df.a_x.round() / df.b_x,
        "a_y": df.a_y.round() / df.b_x
    })

    assert_frame_equal(res, dst)


def test_across_names_arg(df):
    res = across(df, _[_.a_x, _.a_y], Fx + 1, names="{col}_funkyname")
    assert list(res.columns) == ["a_x_funkyname", "a_y_funkyname"]

    dst = (df[["a_x", "a_y"]] + 1).rename(columns = lambda s: s + "_funkyname")
    assert_frame_equal(res, dst)


def test_across_func_dict(df):
    res = across(df, _[_.a_x, _.a_y], {"plus1": Fx + 1, "plus2": Fx + 2})

    dst = pd.DataFrame({
        "a_x_plus1": df.a_x + 1,
        "a_x_plus2": df.a_x + 2,
        "a_y_plus1": df.a_y + 1,
        "a_y_plus2": df.a_y + 2
    })

    assert_frame_equal(res, dst)


def test_across_func_dict_names_arg(df):
    # TODO: also test aggregation
    funcs = {"plus1": Fx + 1, "plus2": Fx + 2}
    res = across(df, _[_.a_x, _.a_y], funcs, names="{fn}_{col}")

    dst = pd.DataFrame({
        "plus1_a_x": df.a_x + 1,
        "plus2_a_x": df.a_x + 2,
        "plus1_a_y": df.a_y + 1,
        "plus2_a_y": df.a_y + 2,
    })

    assert_frame_equal(res, dst)





from siuba.siu import CallTreeLocal, FunctionLookupError, ExecutionValidatorVisitor
from .groupby import SeriesGroupBy

from .translate import (
        forward_method,
        not_implemented,
        method_agg_op,
        method_win_op_agg_result
        )

from siuba.experimental.pd_groups.groupby import SeriesGroupBy, GroupByAgg, broadcast_agg, is_compatible


# THE REAL DIALECT FILE LET'S DO THIS
# ====================================

from siuba.ops import ALL_OPS
from siuba import ops

# register concrete implementations for all ops -------------------------------
# note that this may include versions that would error (e.g. tries to look
# up a Series method that doesn't exist). Custom implementations to fix
# are registered over these further down
for  dispatcher in ALL_OPS.values():#vars(ops).values():
    #try:
    forward_method(dispatcher)
    #except KeyError:
    #    pass


# custom implementations ------------------------------------------------------

def register_method(ns, op_name, f, is_property = False, accessor = None):
    generic = ns[op_name]
    return generic.register(SeriesGroupBy, f(op_name, is_property, accessor))

# add to new op spec
# create new call tree

# aggregate ----

NOT_IMPLEMENTED_AGG = [
        "bool", "dot", "empty", "equals", "hasnans", "is_unique", "kurt",
        "kurtosis", "memory_usage", "nbytes", "product"
        ]

for f_name in NOT_IMPLEMENTED_AGG:
    ALL_OPS[f_name].register(SeriesGroupBy, not_implemented(f_name))

# size is a property on ungrouped, but not grouped pandas data.
# since siuba follows the ungrouped API, it's used as _.x.size, and
# just needs its implementation registered as a *non*-property.
ops.size.register(SeriesGroupBy, method_agg_op("size", is_property = False, accessor = None))


# window ----
NOT_IMPLEMENTED_WIN = [
        "asof", "at", "autocorr", "cat.remove_unused_categories", 
        "convert_dtypes", "drop_duplicates", "duplicated", "get",
        "iat", "iloc", "infer_objects", "is_monotonic",
        ]

for f_name in NOT_IMPLEMENTED_WIN:
    ALL_OPS[f_name].register(SeriesGroupBy, not_implemented(f_name))


# a few functions apply window operations, but return an agg-like result
forward_method(ops.is_monotonic_decreasing, method_win_op_agg_result)
forward_method(ops.is_monotonic_increasing, method_win_op_agg_result)


# NOTE TODO: these methods could be implemented, but depend on the type of 
# time index they're operating on
NOT_IMPLEMENTED_DT = [
    "dt.qyear", "dt.to_pytimedelta", "dt.to_timestamp", "dt.total_seconds", "dt.tz_convert",
    "to_period","dt.to_pydatetime"
    ]

for f_name in NOT_IMPLEMENTED_DT:
    ALL_OPS[f_name].register(SeriesGroupBy, not_implemented(f_name))





# ====================================

from .translate import GroupByAgg, SeriesGroupBy

# TODO: use pandas call tree creator
from siuba.ops.generics import ALL_PROPERTIES, ALL_ACCESSORS

call_listener = CallTreeLocal(
        ALL_OPS,
        call_sub_attr = ALL_ACCESSORS,
        chain_sub_attr = True,
        dispatch_cls = GroupByAgg,
        result_cls = SeriesGroupBy,
        call_props = ALL_PROPERTIES
        )

call_validator = ExecutionValidatorVisitor(GroupByAgg, SeriesGroupBy)


# Fast group by verbs =========================================================

from siuba.siu import Call, singledispatch2
from siuba.dply.verbs import mutate, filter, summarize, DataFrameGroupBy
from pandas.core.dtypes.inference import is_scalar
import warnings

def fallback_warning(expr, reason):
    warnings.warn(
            "The expression below cannot be executed quickly. "
            "Using the slow (but general) pandas apply method."
            "\n\nExpression: {}\nReason: {}"
                .format(expr, reason)
            )


def grouped_eval(__data, expr, require_agg = False):
    if is_scalar(expr):
        return expr
    
    if isinstance(expr, Call):
        try:
            call = call_listener.enter(expr)
            call_validator.visit(call)

        except FunctionLookupError as e:
            fallback_warning(expr, str(e))
            call = expr

        #
        grouped_res = call(__data)

        if isinstance(grouped_res, SeriesGroupBy):
            if not is_compatible(grouped_res, __data):
                raise ValueError("Incompatible groupers")

            # TODO: may want to validate result is correct length / index?
            #       e.g. a SeriesGroupBy could be compatible and not an agg
            if require_agg:
                return grouped_res.obj
            else:
                # broadcast from aggregate to original length (like transform)
                return broadcast_agg(grouped_res)

        else:
            # can happen right now if user selects, e.g., a property of the
            # groupby object, like .dtype, which returns a single value
            # in the future, could restrict set of operations user could perform
            raise ValueError("Result must be subclass of SeriesGroupBy")

    raise ValueError("Grouped expressions must be a siu expression or scalar")



# Fast mutate ----

def _transform_args(args):
    out = []
    for expr in args:
        if is_scalar(expr):
            out.append(expr)
        elif isinstance(expr, Call):
            try:
                call = call_listener.enter(expr)
                call_validator.visit(call)
                out.append(call)
            except FunctionLookupError as e:
                fallback_warning(expr, str(e))
                return None
        elif callable(expr):
            return None

    return out

def _copy_dispatch(dispatcher, cls, func = None):
    if func is None:
        return lambda f: _copy_dispatch(dispatcher, cls, f)

    # Note stripping symbolics may occur twice. Once in the original, and once
    # in this dispatcher.
    new_dispatch = singledispatch2(cls, func)
    new_dispatch.register(object, dispatcher)

    return new_dispatch


@_copy_dispatch(mutate, DataFrameGroupBy)
def fast_mutate(__data, **kwargs):
    """Warning: this function is experimental"""

    # transform call trees, potentially bail out to slow method --------
    new_vals = _transform_args(kwargs.values())

    if new_vals is None:
        return mutate(__data, **kwargs)


    # perform fast method ----
    out = __data.obj.copy()
    groupings = __data.grouper.groupings


    for name, expr in zip(kwargs, new_vals):
        res = grouped_eval(__data, expr)
        out[name] = res

    return out.groupby(groupings)


# Fast filter ----

@_copy_dispatch(filter, DataFrameGroupBy)
def fast_filter(__data, *args):
    """Warning: this function is experimental"""

    # transform call trees, potentially bail out to slow method --------
    new_vals = _transform_args(args)

    if new_vals is None:
        return filter(__data, *args)

    # perform fast method ----
    out = []
    groupings = __data.grouper.groupings

    for expr in args:
        res = grouped_eval(__data, expr)
        out.append(res)

    filter_df = filter.dispatch(__data.obj.__class__)

    df_result = filter_df(__data.obj, *out)

    # TODO: research how to efficiently & robustly subset groupings
    group_names = [ping.name for ping in groupings]
    return df_result.groupby(group_names)


# Fast summarize ----

@_copy_dispatch(summarize, DataFrameGroupBy)
def fast_summarize(__data, **kwargs):
    """Warning: this function is experimental"""

    # transform call trees, potentially bail out to slow method --------
    new_vals = _transform_args(kwargs.values())

    if new_vals is None:
        return summarize(__data, **kwargs)

    # perform fast method ----
    groupings = __data.grouper.groupings

    # TODO: better way of getting this frame?
    out = __data.grouper.result_index.to_frame()
    
    for name, expr in kwargs.items():
        # special case: set scalars directly
        res = grouped_eval(__data, expr, require_agg = True)

        out[name] = res

    return out.reset_index(drop = True)





"""
Note: this test file was heavily influenced by its dbplyr counterpart.

https://github.com/tidyverse/tidyr/blob/master/tests/testthat/test-pivot-long.R
"""

# TODO: need to be careful about index. Should preserve original indices for rows.
# TODO: test when original input has a multi-index (should be okay)

from . import pivot_longer, pivot_longer_spec, build_longer_spec

import pytest
import pandas as pd
import numpy as np

from siuba.siu import Symbolic
from siuba import group_by, collect, show_query

from siuba.tests.helpers import data_frame, assert_equal_query, assert_frame_sort_equal
from pandas.testing import assert_frame_equal, assert_series_equal


_ = Symbolic()



def assert_equal_query2(left, right, *args, sql_ordered=False, sql_kwargs=None, **kwargs):
    from siuba.sql import LazyTbl
    from siuba.sql.utils import _is_dialect_duckdb
    from pandas.core.groupby import DataFrameGroupBy
    out = collect(left)

    if isinstance(left, LazyTbl) and _is_dialect_duckdb(left.source):
        # TODO: find a nice way to remove duckdb specific code from here
        # duckdb does not use pandas.DataFrame.to_sql method, which coerces
        # everything to 64 bit. So we need to coerce any results it returns
        # as 32 bit to 64 bit, to match to_sql.
        int_cols = out.select_dtypes('int32').columns
        flt_cols = out.select_dtypes('float32').columns
        out[int_cols] = out[int_cols].astype('int64')
        out[flt_cols] = out[flt_cols].astype('float64')

    if isinstance(left, LazyTbl):
        if sql_kwargs:
            final_kwargs = {**kwargs, **sql_kwargs}
        else:
            final_kwargs = kwargs
        if not sql_ordered:
            assert_frame_sort_equal(out, right, *args, **final_kwargs)
        else:
            assert_frame_equal(out, right, *args, **final_kwargs)
    else:
        assert_frame_equal(out, right, *args, **kwargs)


def assert_series_equal2(left, right, *args, sql_ordered=True, **kwargs):
    from siuba.sql import LazyTbl

    if isinstance(left, LazyTbl):
        left = collect(left)

        if not sql_ordered:
            left = left.sort_values()
            right = right.sort_values()

        assert_series_equal(
            left.reset_index(drop=True),
            right.reset_index(drop=True),
            *args,
            **kwargs
        )
    else:
        assert_series_equal(left, right, *args, **kwargs)


    
def test_pivot_all_cols_to_long(backend):
    "can pivot all cols to long"

    src = backend.load_df(data_frame(x = [1,2], y = [3,4]))
    dst = data_frame(name = ["x", "y", "x", "y"], value = [1, 3, 2, 4])
    
    query = pivot_longer(_, _["x":"y"])

    assert_equal_query(src, query, dst)


def test_values_interleaved_correctly():
    df = data_frame(x = [1,2], y = [10, 20], z = [100, 200])

    pv = pivot_longer(df, _[0:3])
    assert pv["value"].tolist() == [1, 10, 100, 2, 20, 200]


def test_spec_add_multi_columns():
    # TODO: SQL backends
    df = data_frame(x = [1,2], y = [3,4])

    sp = data_frame(**{".name": ["x", "y"], ".value": "v", "a": 1, "b": 2})
    pv = pivot_longer_spec(df, spec = sp)

    assert pv.columns.tolist() == ["a", "b", "v"]


def test_preserves_original_keys(backend):
    src = data_frame(x = [1,2], y = [2,2], z = [1,2])
    dst = data_frame(x = [1, 1, 2, 2], name = ["y", "z"]*2, value = [2,1,2,2])
    dst.index = [0, 0, 1, 1]

    remote = backend.load_df(src)
        
    pv = pivot_longer(remote, _["y":"z"])

    assert_equal_query2(pv, dst)


def test_can_drop_missing_values(backend):
    src = backend.load_df(
        data_frame(x = [1, np.nan], y = [np.nan, 2])
    )
    dst = data_frame(name = ["x", "y"], value = [1, 2])
    pv = pivot_longer(src, _["x":"y"], values_drop_na=True)

    # TODO: sql databases sometimes return float. we need assertions to take
    # backend (or sql vs pandas) specific options.
    assert_equal_query2(pv, dst, check_dtype=False)


def test_can_handle_missing_combinations(backend):
    df = data_frame(id = ["A", "B"], x_1 = [1, 3], x_2 = [2, 4], y_2 = ["a", "b"])
    dst = data_frame(
        id = ["A", "A", "B", "B"],
        n = ["1", "2"]*2,
        x = [1, 2, 3, 4],
        y = [np.nan, "a", np.nan, "b"]
    )
    dst.index = [0, 0, 1, 1]

    src = backend.load_df(df)
    pv = pivot_longer(src, -_.id, names_to = (".value", "n"), names_sep = "_")

    assert_equal_query2(pv, dst)


@pytest.mark.xfail
def test_mixed_columns_are_auto_coerced():
    # TODO: pandas stack (and melt) coerces categorical data when stacking.
    df = data_frame(x = pd.Categorical(["a"]), y = pd.Categorical(["b"]))
    pv = pivot_longer(df, _["x":"y"])

    assert_series_equal(pv["value"], pd.Categorical(['a', 'b']))


def test_can_override_default_output_col_type():
    df = data_frame(x = "x", y = 1)
    pv = pivot_longer(df, _["x":"y"], values_transform = {"value": lambda x: x.astype("str")})

    assert pv["value"].tolist() == ["x", "1"]


def test_spec_can_pivot_to_multi_measure_cols(backend):
    df = backend.load_df(data_frame(x = "x", y = 1))
    sp = data_frame(**{".name": ["x", "y"], ".value": ["X", "Y"], "row": [1, 1]})

    pv = collect(pivot_longer_spec(df, sp))

    assert pv.columns.tolist() == ["row", "X", "Y"]
    assert pv["X"].tolist() == ["x"]
    assert pv["Y"].tolist() == [1]


def test_original_col_order_is_preserved(backend):
    df = data_frame(
        id = ["A", "B"],
        z_1 = [1, 7], y_1 = [2, 8], x_1 = [3, 9],
        z_2 = [4, 10], y_2 = [5, 11], x_2 = [6, 12]
    )
    src = backend.load_df(df)
    pv = collect(
        pivot_longer(src, -_.id, names_to = (".value", "n"), names_sep = "_")
    )

    assert pv.columns.tolist() == ["id", "n", "z", "y", "x"]


def test_handles_duplicate_column_names():
    # Cannot initiate data_frame with duplicate keys
    # df = data_frame(x = 1, a = 1, a = 2, b = 3, b = 4)
    df = pd.DataFrame.from_records(
        [(1, 1, 2, 3, 4)], columns = ["x", "a", "a", "b", "b"]
    )
    pv = pivot_longer(df, -_.x)

    assert pv.columns.tolist() == ["x", "name", "value"]
    assert pv["name"].tolist() == ["a", "a", "b", "b"]
    assert pv["value"].tolist() == [1, 2, 3, 4]


def test_can_pivot_duplicate_names_to_value(backend):
    df = data_frame(x = 1, a_1 = 1, a_2 = 2, b_1 = 3, b_2 = 4)
    src = backend.load_df(df)

    pv1 = pivot_longer(df, -_.x, names_to = (".value", np.nan), names_sep = "_")
    pv2 = pivot_longer(df, -_.x, names_to = (".value", np.nan), names_pattern = "(.)_(.)")
    pv3 = pivot_longer(df, -_.x, names_to = ".value", names_pattern = "(.)_.")

    assert pv1.columns.tolist() == ["x", "a", "b"]
    assert pv1["a"].tolist() == [1, 2]
    assert_frame_equal(pv2, pv1)
    assert_frame_equal(pv3, pv1)


def test_value_can_be_any_pos_in_names_to():
    # TODO: should work in SQL
    samp = data_frame(
        i = np.arange(1, 5),
        y_t1 = np.random.standard_normal(4),
        y_t2 = np.random.standard_normal(4),
        z_t1 = [3] * 4,
        z_t2 = [-2] * 4,
    )

    value_first = pivot_longer(samp, -_.i,
                               names_to = (".value", "time"), names_sep = "_")

    samp2 = samp.rename(columns={"y_t1": "t1_y", "y_t2": "t2_y",
                                 "z_t1": "t1_z", "z_t2": "t2_z"})
    
    value_second = pivot_longer(samp2, -_.i,
                                names_to = ("time", ".value"), names_sep = "_")
    
    assert_frame_equal(value_first, value_second)


@pytest.mark.skip("TODO")
def test_type_error_message_uses_var_names():
    """type error message use variable names"""
    # Error handling is by default 'better' in python than R
    # This test is tricky, as python doesn't care when stacking data of different types.
    df = data_frame(abc = 1, xyz = "b")
    try:
        # This should by default pivot everything, as with tidyr
        pivot_longer(df, _[:])
    except:
        # Ideally we'd print an error message here and compare if the keys are
        # printed correctly, but `pivot_longer` doesn't brake in python as in R
        # when stacking different data types. Not sure if we should 'make' it brake?
        print(err)


@pytest.mark.skip("wontdo?")
def test_values_ptypes_raises_variable_name():
    assert False


@pytest.mark.skip("wontdo?")
def test_names_ptypes_raises_names_to_names():
    assert False


def test_error_overwriting_existing_column():
    src = data_frame(x = 1, y = 2)

    # TODO: test exact message
    with pytest.raises(ValueError) as exc_info:
        pivot_longer(src, _.y, names_to = "x")

    pv = pivot_longer(src, _.y, names_to = "x", names_repair="unique")
    assert_frame_equal(
        pv,
        data_frame(x___0=1, x___1="y", value=2)
    )

def test_grouping_is_preserved():
    # In siuba this actually tests 3 things:
    # 1) Can we pipe a grouped DataFrame in to `pivot_longer`? - Not yet
    # 2) Does it retain grouping? - Not yet
    # 3) Can we get the names of grouping _columns_ ie variables? - Not yet?
    df = data_frame(g = [1, 2], x1 = [1, 2], x2 = [3, 4])
    # Breaks; `pivot_longer` needs singledispatch for grouped DataFrames
    out = (
        df
        >> group_by(_.g)
        >> pivot_longer(_["x1":"x2"], names_to = "x", values_to = "v")
    )

    # Breaks, as group_vars does not exist yet.
    # For now, in pandas it is probably better to check if the DataFrame remains
    # grouped, and if it matches the expected output
    # assert group_vars(out) == "g"
    expected = data_frame(
        g = [1, 1, 2, 2],
        x = ["x1", "x2", "x1", "x2"],
        v = [1, 3, 2, 4],
        _index = [0, 0, 1, 1]
    ).groupby("g")

    # assert_frame_equal does not work with DataFrameGroupBy.
    isinstance(out, expected.__class__)
    assert_frame_equal(out.obj, expected.obj)


def test_zero_row_frames_work():
    src = pd.DataFrame({"x": [], "y": []}, dtype=int)
    pv = pivot_longer(src, _[_.x, _.y])

    assert_frame_equal(
        pv,
        pd.DataFrame({
            "name": np.array([], dtype="object"),
            "value": np.array([], dtype="int")
        })
    )


@pytest.mark.skip("unreleased in tidyr")
def test_cols_vary_adjust_result_row_order():
    src = data_frame(x = [1, 2], y = [3, 4])

    assert_frame_equal(
        pivot_longer(src, _[_.x, _.y], cols_vary = "fastest"),
        data_frame(name = ["x", "y", "x", "y"], value = [1, 3, 2, 4])
    )

    assert_frame_equal(
        pivot_longer(src, _[_.x, _.y], cols_vary = "slowest"),
        data_frame(name = ["x", "x", "y", "y"], value = [1, 2, 3, 4])
    )


@pytest.mark.skip("unreleased in tidyr")
def test_cols_vary_works_with_id_cols():
    """"`cols_vary` works with id columns not part of the pivoting process"""
    src = data_frame(id = ["a", "b"], x = [1, 2], y = [3, 4])

    pv = pivot_longer(src, _[_.x, _.y], cols_vary = "fastest")

    assert list(pv["id"]) == ["a", "a", "b", "b"]
    assert_frame_equal(
        pv[["name", "value"]],
        pivot_longer(src[["x", "y"]], _[_.x, _.y], cols_vary = "fastest")
    )


    pv2 = pivot_longer(src, _[_.x, _.y], cols_vary = "slowest")

    assert pv2["id"] == ["a", "b", "a", "b"]
    assert_frame_equal(
        pv2,
        pivot_longer(src[["x", "y"]], _[_.x, _.y], cols_vary = "slowest")
    )


@pytest.mark.skip("unreleased in tidyr")
def test_cols_vary_works_with_values_drop_na():
    src = data_frame(id = ["a", "b"], x = [1, None], y = [3, 4])

    assert_frame_equal(
        pivot_longer(src, _[_.x, _.y], cols_vary="slowest", values_drop_na = True),
        data_frame(id = ["a", "a", "b"], name = ["x", "y", "y"], value = [1, 3, 4])
    )


# spec ------------------------------------------------------------------------

def test_spec_validates_input():
    src = data_frame(x = 1)

    # TODO: more specific check: class = "vctrs_error_assert"
    with pytest.raises(Exception):
        build_longer_spec(src, _.x, values_to = ["a", "b"])


@pytest.mark.xfail
def test_names_to_empty_no_names_in_result():
    src = data_frame(x = 1)

    assert_frame_equal(
        build_longer_spec(src, _.x, names_to = []),
        spec,
        data_frame({".name": "x", ".value": "value"})
    )

    assert_frame_equal(
        build_longer_spec(src, names_to = None),
        spec
    )


    # TODO: test what happens when you pivot this? should result in...
    # data_frame(x = 1)


@pytest.mark.xfail
def test_multiple_names_requires_name_sep_or_pattern():
    """multiple names requires names_sep/names_pattern"""

    src = data_frame(x_y = 1)

    # TODO: assert this raises an error
    build_longer_spec(src, _.x_y, names_to = c("a", "b"))

    # TODO: assert this raises an error
    build_longer_spec(src, _.x_y, names_to = c("a", "b"), names_sep = "x", names_pattern = "x")


def test_spec_names_sep():
    src = data_frame(x_y = 1)
    spec = build_longer_spec(src, _.x_y, names_to = ["a", "b"], names_sep = "_")

    assert_frame_equal(
        spec,
        data_frame(**{".name": "x_y", ".value": "value", "a": "x", "b": "y"})
    )


def test_spec_names_sep_fails_single_name():
    src = data_frame(x_y = 1)

    # TODO: better test of exception
    with pytest.raises(Exception):
        build_longer_spec(df, _.x_y, names_to = "x", names_sep = "_")


def test_names_pattern_generates_correct_spec():
    src = data_frame(zx_y = 1)
    spec = build_longer_spec(
        src,
        _.zx_y,
        names_to = ["a", "b"],
        names_pattern = "z(.)_(.)"
    )

    dst = data_frame(**{".name": "zx_y", ".value": "value", "a": "x", "b": "y"})
    assert_frame_equal(
        spec,
        dst
    )

    spec2 = build_longer_spec(src, _.zx_y, names_to = "a", names_pattern = "z(.)")
    assert_frame_equal(
        spec2,
        dst.drop(columns="b")
    )


def test_names_to_can_override_value_to():
    src = data_frame(x_y = 1)
    spec = build_longer_spec(src, _.x_y, names_to = ["a", ".value"], names_sep = "_")

    assert_frame_equal(
        spec,
        # TODO: tidyr puts .value after a -- is that a bug?
        data_frame(**{".name": "x_y", ".value": "y", "a": "x"})
    )


def test_names_prefix_strips_off_from_beginning():
    src = data_frame(zzyz = 1)
    spec = build_longer_spec(src, _.zzyz, names_prefix = "z")

    assert_frame_equal(
        spec,
        data_frame(**{".name": "zzyz", ".value": "value", "name": "zyz"})
    )


def test_spec_can_cast_to_custom_type():
    src = data_frame(w1 = 1)
    spec = build_longer_spec(
        src,
        _.w1,
        names_prefix = "w",
        names_transform = {"name": lambda x: x.astype(int)}
    )

    assert spec["name"].dtype == "int64"


@pytest.mark.skip("wontdo? uses ptype")
def test_spec_transform_applied_before_cast():
    src = data_frame(w1 = 1)
    assert False


@pytest.mark.skip("wontdo? uses ptype")
def test_names_ptype_works_with_single_value():
    assert False


def test_names_transform_works_with_single_value():
    src = data_frame(**{"1x2": 1})
    spec = build_longer_spec(
        src,
        _["1x2"],
        names_to = ["one", "two"],
        names_sep = "x",
        names_transform = lambda x: x.astype(int)
    )

    assert list(spec["one"]) == [1]
    assert list(spec["two"]) == [2]


@pytest.mark.skip("wontdo? ptype")
def test_values_ptypes_works_with_single_empty_ptypes():
    assert False


def test_values_transform_works_with_single_functions():
    src = data_frame(x_1 = 1, y_1 = 2)
    pv = pivot_longer(
        src, _[:],
        names_to = [".value", "set"],
        names_sep = "_",
        values_transform = lambda x: x.astype(str)
    )


@pytest.mark.skip("wontdo? ptypes")
def test_names_and_values_ptypes_empty_dict():
    """`names/values_ptypes = list()` is currently the same as `NULL` (#1296)"""
    assert False


def test_error_if_col_no_match():
    src = data_frame(x = 1, y = 2)

    # TODO: better exception handling
    with pytest.raises(Exception):
        pivot_longer(src, _.startswith("z"))


def test_names_to_is_validated():
    src = data_frame(x = 1)

    # TODO: better exception testing
    with pytest.raises(Exception):
        build_longer_spec(src, _.x, names_to = 1)

    with pytest.raises(Exception):
        build_longer_spec(src, _.x, names_to = ["x", "y"])

    with pytest.raises(Exception):
        build_longer_spec(src, _.x, names_to = ["x", "y"], names_sep="_", names_pattern="x")
    

@pytest.mark.skip("wontdo? uses ptypes")
def test_that_names_ptypes_is_validated():
    assert False


def test_names_transform_is_validated():
    src = data_frame(x = 1)

    with pytest.raises(Exception):
        build_longer_spec(src, _.x, names_transform = 1)

    with pytest.raises(Exception):
        build_longer_spec(src, _.x, names_transform = [lambda x: x])


@pytest.mark.skip("wontdo? ptypes")
def test_values_ptypes_is_validated():
    assert False


def test_values_transform_is_validated():
    src = data_frame(x = 1)
    
    with pytest.raises(Exception):
        pivot_longer(src, _.x, values_transform = 1)

    with pytest.raises(Exception):
        pivot_longer(src, _.x, values_transform = [lambda x: x])


def test_cols_vary_is_validated():
    src = data_frame(x = 1)

    with pytest.raises(Exception):
        pivot_longer(src, _.x, cols_vary = "fast")

    with pytest.raises(Exception):
        pivot_longer(src, _.x, cols_vary = 1)



import pandas as pd

from sqlalchemy import sql
from sqlalchemy import types as sqla_types

from siuba.dply.verbs import spread
from siuba.sql import LazyTbl
from siuba.sql.utils import (
    _sql_select,
    _sql_column_collection,
    _sql_add_columns,
    _sql_with_only_columns,
)


from .pivot_long import pivot_longer, pivot_longer_spec, build_longer_spec, spec_to_multiindex


def unpack_spec_row(d):
    internal = {".name", ".value"}
    return d[".name"], d[".value"], {k:v for k,v in d.items() if k not in internal}


def _safe_to_dict(df, *args, **kwargs):
    """Return something like df.to_dict(), but ensure t contains no numpy types.

    For context on this pandas issue, see issues linked in this PR:
    https://github.com/pandas-dev/pandas/issues/13258

    """
    import json
    return json.loads(df.to_json(*args, **kwargs))


def _values_to_select(sel_columns, spec_row: dict, value_vars: "list[str]"):
    final_cols = []
    for long_name in value_vars:
        wide_name = spec_row[long_name]
        if pd.isna(wide_name):
            final_cols.append(sql.null().label(long_name))
        else:
            final_cols.append(sel_columns[wide_name].label(long_name))

    return final_cols


@build_longer_spec.register
def _build_longer_spec(__data: LazyTbl, *args, **kwargs):
    # building the spec only really needs the columns names. however, because we 
    # matched tidyr behavior, we just pass a DataFrame in for now.
    df_data = pd.DataFrame(columns = list(__data.last_op.alias().columns.keys()))

    return build_longer_spec(df_data, *args, **kwargs)


@pivot_longer_spec.register
def _pivot_longer_spec(
    __data: LazyTbl,
    spec,
    names_repair = "check_unique",
    values_drop_na: bool = False,
    values_ptypes = None,
    values_transform = None
) -> LazyTbl:

    if values_ptypes is not None:
        raise NotImplementedError()

    if values_transform is not None:
        raise NotImplementedError()


    sel = __data.last_op
    sel_alias = sel.alias()


    # extract info from spec ----

    column_index = spec_to_multiindex(spec)

    wide_names = list(spec[".name"].unique())
    wide_ids = [name for name in sel_alias.columns.keys() if name not in wide_names]

    long_name_vars = [k for k in spec.columns if k not in {".name", ".value"}]
    long_val_vars = list(spec[".value"].unique())


    # guard against bad specifications ----

    bad_names = set(wide_names) - set(sel_alias.columns.keys())
    if bad_names:
        raise ValueError(f"Pivot spec contains columns not in the data: {bad_names}")


    # reshape to long (via union all) ----

    sel_cols = sel_alias.columns

    # each row maps <new_name>: literal for name vars, or <new_name>: column 
    aligned_vars = spread(spec, ".value", ".name")

    union_parts = []
    for row in _safe_to_dict(aligned_vars, orient="records"):
        id_cols = [sel_cols[_id] for _id in wide_ids]
        
        # TODO: handle when value name (row[k]) is NULL
        value_cols = _values_to_select(sel_cols, row, long_val_vars)
        name_cols = [
            sql.literal(row[k]).label(k)
            for k in long_name_vars
        ]

        union_parts.append(_sql_select([*id_cols, *name_cols, *value_cols]))

    # TODO: what is the base class we are willing to let the select type be?
    # this is a CompoundSelect. Shares GenerativeSelect with sql.select()
    sel_union = sql.union_all(*union_parts)

    if values_drop_na:
        alias = sel_union.alias()

        # TODO: sqlalchemy 1.4+ prefers .is_not()
        bool_clause = sql.and_(*[alias.columns[k].isnot(None) for k in long_val_vars])

        return __data.append_op(alias.select().where(bool_clause))

    return __data.append_op(sel_union)
        

# simply calls build_longer_spec and pivot_longer_spec
pivot_longer.register(LazyTbl, pivot_longer.dispatch(pd.DataFrame))






"""
Implements LazyTbl to represent tables of SQL data, and registers it on verbs.

This module is responsible for the handling of the "table" side of things, while
translate.py handles translating column operations.


"""

from sqlalchemy import sql

from siuba.dply.verbs import count, add_count, inner_join, _check_name

from ..utils import _sql_select, _sql_add_columns, lift_inner_cols
from ..backend import LazyTbl, ordered_union
from ..translate import AggOver

from .mutate import _mutate_cols


@count.register(LazyTbl)
def _count(__data, *args, sort = False, wt = None, name=None, **kwargs):
    if wt is not None:
        raise NotImplementedError("wt argument is currently not implemented")

    result_names, sel_inner = _mutate_cols(__data, args, kwargs, "Count")

    # remove unnecessary select, if we're operating on a table ----
    if set(lift_inner_cols(sel_inner)) == set(lift_inner_cols(__data.last_select)):
        sel_inner = __data.last_op

    # create outer select ----
    # holds selected columns and tally (n)
    sel_inner_cte = sel_inner.alias()
    inner_cols = sel_inner_cte.columns

    # apply any group vars from a group_by verb call first
    missing = [k for k in __data.group_by if k not in result_names]

    all_group_names = ordered_union(__data.group_by, result_names)
    outer_group_cols = [inner_cols[k] for k in all_group_names]

    # holds the actual count (e.g. n)
    label_n = _check_name(name, set(inner_cols.keys()))
    count_col = sql.functions.count().label(label_n)

    sel_outer = _sql_select([*outer_group_cols, count_col]) \
            .select_from(sel_inner_cte) \
            .group_by(*outer_group_cols)

    # count is like summarize, so removes order_by
    return __data.append_op(
            sel_outer.order_by(count_col.desc()),
            order_by = tuple()
            )


@add_count.register(LazyTbl)
def _add_count(__data, *args, wt = None, sort = False, name=None, **kwargs):
    if wt is not None:
        raise NotImplementedError("wt argument is currently not implemented")

    result_names, sel_inner = _mutate_cols(__data, args, kwargs, "Count")

    # TODO: if clause copied from count
    # remove unnecessary select, if we're operating on a table ----
    if set(lift_inner_cols(sel_inner)) == set(lift_inner_cols(__data.last_select)):
        sel_inner = __data.last_select

    inner_cols = lift_inner_cols(sel_inner)


    # TODO: this code to append groups to columns copied a lot inside verbs
    # apply any group vars from a group_by verb call first
    missing = [k for k in __data.group_by if k not in result_names]

    all_group_names = ordered_union(__data.group_by, result_names)
    outer_group_cols = [inner_cols[k] for k in all_group_names]


    count_col = AggOver(sql.functions.count(), partition_by=outer_group_cols)
    label_n = _check_name(name, set(inner_cols.keys()))

    sel_appended = _sql_add_columns(sel_inner, [count_col.label(label_n)])

    return __data.append_op(sel_appended)



from contextlib import contextmanager, AbstractContextManager
import sys
import copy

ERROR_INSTRUCTIONS = "To see the full error, run: import siuba.siu.error; raise siuba.siu.error.last()"

def last():
    """Return last error with suppress context set to False.
    
    Note: makes a copy of last error, since interactive shells like ipython
          do later handling of exceptions, so a context manager is not an option.

    """

    err = sys.last_value

    if err is None:
        return None

    err.__suppress_context__ = False
    return err


class ShortException(Exception):
    """Exception that provides the option (short) to remove context from stack traces.
    
    This greatly shortens the stack. Instructions are provided for viewing the whole thing.

    """
    def __init__(self, msg, short = False):
        if short: 
            full_msg = str(msg) + "\n\n" + ERROR_INSTRUCTIONS
        else:
            full_msg = msg

        self.args = (full_msg,)
        self.__suppress_context__ = short


    @classmethod
    def from_verb(cls, verb_name, arg_name, err, *args, **kwargs):
        error_msg = "\nVerb: {verb}\nArg: {arg}\nError: {err}.".format(
                verb = verb_name, arg = arg_name, err = err
                )
        
        return cls(error_msg, *args, **kwargs)




import pytest

from siuba.siu.calls import PipeCall
from siuba.siu.dispatchers import call
from siuba.siu import _, strip_symbolic, Symbolic

# TODO: direct test of lazy elements
# TODO: NSECall - no map subcalls


def test_siu_pipe_call_is_flat():
    pipe_expr = _ >> _.a >> _.b
    pipe_call = strip_symbolic(pipe_expr)

    assert isinstance(pipe_expr, Symbolic)
    assert isinstance(pipe_call, PipeCall)
    assert len(pipe_call.args) == 4

def test_siu_call_no_args():
    assert 1 >> call(range) == range(1)

def test_siu_call_no_args_explicit():
    assert 1 >> call(range, _) == range(1)

def test_siu_call_pos_arg():
    assert 1 >> call(range, _, 2) == range(1, 2)


def test_siu_call_kwarg():
    assert "," >> call("a,b,c".split, _, maxsplit=1) == ["a", "b,c"]


def test_siu_call_onlykwargs():
    def f(*, x):
        return x

    res = 1 >> call(f, x = _)
    assert res == 1


def test_siu_call_arg_kwarg():
    assert 1 >> call("{0}_{1}_{b}".format, _, 2, b=3) == "1_2_3"


def test_siu_call_underscore_arg():
    assert 1 >> call(range, 2, 3, _) == range(2, 3, 1)


def test_siu_call_underscore_method():
    assert "a,b" >> call(_.split(",")) == ["a", "b"]


def test_siu_call_underscore_method_args():
    with pytest.raises(NotImplementedError):
        "a,b" >> call(_.split, _, ",")



"""
Translations for the postgresql dialect of SQL.
"""

from ..translate import (
        win_agg, win_over, win_cumul, sql_scalar, sql_agg, win_absent,
        RankOver,
        wrap_annotate, annotate,
        extend_base,
        SqlTranslator,
        )

from .base import (
        SqlColumn, SqlColumnAgg
        )

import sqlalchemy.sql.sqltypes as sa_types
from sqlalchemy import sql
from sqlalchemy.sql import func as fn


# Custom dispatching in call trees ============================================

class PostgresqlColumn(SqlColumn): pass

class PostgresqlColumnAgg(SqlColumnAgg, PostgresqlColumn): pass

# Custom translations =========================================================

# datetime

@annotate(return_type="float")
def sql_is_quarter_end(_, col):
    last_day = fn.date_trunc("quarter", col) + sql.text("INTERVAL '3 month' - INTERVAL '1 day'")
    return fn.date_trunc("day", col) == last_day



# other

def returns_float(func_names):
    # TODO: MC-NOTE - shift all translations to directly register
    # TODO: MC-NOTE - make an AliasAnnotated class or something, that signals
    #                 it is using another method, but w/ an updated annotation.
    from siuba.ops import ALL_OPS
    
    for name in func_names:
        generic = ALL_OPS[name]
        f_concrete = generic.dispatch(SqlColumn)
        f_annotated = wrap_annotate(f_concrete, result_type="float")
        generic.register(PostgresqlColumn, f_annotated)
    

def sql_log(_, col, base = None):
    if base is None:
        return sql.func.ln(col)
    return sql.func.log(col)

@annotate(result_type = "float")
def sql_round(_, col, n):
    return sql.func.round(col, n)

def sql_func_contains(_, col, pat, case = True, flags = 0, na = None, regex = True):
    # TODO: warn there differences in regex for python and sql?
    # TODO: validate pat is string?
    if not isinstance(pat, str):
        raise TypeError("pat argument must be a string")
    if flags != 0 or na is not None:
        raise NotImplementedError("flags and na options not supported")

    if not regex:
        case_col = col if case else col.lower()
        return case_col.contains(pat, autoescape = True)

    full_op = "~" if case else "~*"

    return col.op(full_op)(pat)

def sql_func_truediv(_, x, y):
    return sql.cast(x, sa_types.Float()) / y


extend_base(
        PostgresqlColumn,

        # TODO: remove log, not a pandas method
        #log = sql_log,

        # TODO: bring up to date (not pandas methods)
        #concat = lambda col: sql.func.concat(col),
        #cat = lambda col: sql.func.concat(col),
        #str_c = lambda col: sql.func.concat(col),

        # infix and infix methods ----

        div = sql_func_truediv,
        divide = sql_func_truediv,
        rdiv = lambda _, x,y: sql_func_truediv(_, y, x),

        __truediv__ = sql_func_truediv,
        truediv = sql_func_truediv,
        __rtruediv__ = lambda _, x, y: sql_func_truediv(_, y, x),


        round = sql_round,
        __round__ = sql_round,

        **{
            "str.contains": sql_func_contains,
        },
        **{
            "dt.is_quarter_end": sql_is_quarter_end
        }
        )

returns_float([
             "dt.day", "dt.dayofweek", "dt.dayofyear", "dt.days_in_month",
             "dt.daysinmonth", "dt.hour", "dt.minute", "dt.month",
             "dt.quarter", "dt.second", "dt.week", "dt.weekday",
             "dt.weekofyear", "dt.year"
             ])

extend_base(
        PostgresqlColumn,
        any = annotate(win_agg("bool_or"), input_type = "bool"),
        all = annotate(win_agg("bool_and"), input_type = "bool"),
        #lag = win_agg("lag"),
        std = win_agg("stddev_samp"),
        var = win_agg("var_samp"),

        # overrides ----

        # note that postgres does sum(bigint) -> numeric
        quantile = win_absent("percentile_cont"),
        size = win_agg("count"),     #TODO double check
        )

extend_base(
        PostgresqlColumnAgg,
        all = sql_agg("bool_and"),
        any = sql_agg("bool_or"),
        std = sql_agg("stddev_samp"),
        var = sql_agg("var_samp"),
        )


# translate(config, CallTreeLocal, PostgresqlColumn, _.a + _.b)
translator = SqlTranslator.from_mappings(
        PostgresqlColumn, PostgresqlColumnAgg
        )



from siuba.dply.verbs import simple_varname
from siuba import _, filter, group_by, arrange, mutate, ungroup
from siuba.dply.vector import row_number, desc
import pandas as pd

import pytest

from .helpers import assert_equal_query, data_frame, backend_notimpl, backend_sql

DATA = data_frame(x = [2,2,1], y = [2,1,1], z = ['z']*3)

@pytest.fixture(scope = "module")
def df(backend):
    return backend.load_df(DATA)


@pytest.mark.parametrize("query, output", [
    (arrange(_.x), DATA.sort_values(['x'])),
    (arrange("x"), DATA.sort_values(['x'])),
    (arrange(_.x, _.y), DATA.sort_values(['x', 'y'])),
    (arrange("x", "y"), DATA.sort_values(['x', 'y'])),
    (arrange(_.x, "y"), DATA.sort_values(['x', 'y']))
    ])
def test_basic_arrange(df, query, output):
    assert_equal_query(df, query, output)


@pytest.mark.parametrize("query, output", [
    (arrange(-_.x), DATA.sort_values(['x'], ascending = [False])),
    (arrange(-_.x, _.y), DATA.sort_values(['x', 'y'], ascending = [False, True])),
    (arrange(-_.x, "y"), DATA.sort_values(['x', 'y'], ascending = [False, True]))
    ])
def test_arrange_desc(df, query, output):
    assert_equal_query(df, query, output)


@pytest.mark.parametrize("query, output", [
    (arrange(_.x - _.x), DATA),
    (arrange(_.x * -1), DATA.sort_values(['x'], ascending = [False])),
    ])
def test_arrange_with_expr(df, query, output):
    assert_equal_query(df, query, output)


def test_arrange_grouped_trivial(df):
    # note: only 1 level for z
    assert_equal_query(
            df,
            group_by(_.z) >> arrange(_.x),
            DATA.sort_values(['x'])
            )

def test_arrange_grouped(backend, df):
    q = group_by(_.y) >> arrange(_.x)
    assert_equal_query(
            df,
            q,
            DATA.sort_values(['x'])
            )

    # arrange w/ mutate is the same, whether used before or after group_by
    assert_equal_query(
            df,
            q >> mutate(res = row_number(_)),
            mutate(DATA.sort_values(['x']).groupby('y'), res = row_number(_))
            )


# SQL -------------------------------------------------------------------------

@backend_sql
def test_no_arrange_before_cuml_window_warning(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)
    with pytest.warns(RuntimeWarning):
        dfs >> mutate(y = _.x.cumsum())

@backend_sql
def test_arranges_back_to_back(backend):
    data = data_frame(x = range(1, 5), g = [1,1,2,2])
    dfs = backend.load_df(data)

    lazy_tbl = dfs >> arrange(_.x) >> arrange(_.g)
    order_by_vars = tuple(simple_varname(call) for call in lazy_tbl.order_by)

    assert order_by_vars == ("g",)
    assert [c.name for c in lazy_tbl.last_op._order_by_clause] == ["g"]





# TODO: this is to make sure everything can be wired up. Should refactor
#       CallTreeLocal to be less finicky.
# NOTE: put at bottom to avoid issue with circular import
#       could be moved into separate module
from .generics import ALL_OPS, ALL_ACCESSORS, ALL_PROPERTIES
from siuba.siu import CallTreeLocal

# TODO: I hate this
def create_pandas_translator(local, dispatch_cls, result_cls):
    return CallTreeLocal(
            local,
            call_sub_attr = tuple(ALL_ACCESSORS),
            chain_sub_attr = True,
            dispatch_cls = dispatch_cls,
            result_cls = result_cls,
            call_props = tuple(ALL_PROPERTIES)
            )




from siuba.siu import strip_symbolic, FunctionLookupError, Symbolic, MetaArg, Call
import pytest

BINARY_OPS = [
    "+",
    "-",
    "*",
    "@",
    "//",
    #truediv
    #divmod
    "/",
    "%",
    #"_ ** _", #TODO: uses different formatting
    "<<",
    #">>",
    "&",
    "^",
    "|"
    ]

@pytest.fixture
def _():
    return Symbolic()


# Symbolic class ==============================================================

def test_symbolic_source_attr(_):
    sym = _.__source
    assert isinstance(sym, Symbolic)
    assert explain(sym) == "_.__source"


def test_symbolic_numpy_ufunc(_):
    from siuba.siu.symbolic import array_ufunc
    import numpy as np
    
    # should have form...
    # █─'__call__'
    # ├─█─'__custom_func__'
    # │ └─<function array_ufunc at ...>
    # ├─_
    # ├─<ufunc 'sqrt'>
    # ├─'__call__'
    # └─_

    sym = np.sqrt(_)
    expr = strip_symbolic(sym)

    assert isinstance(sym, Symbolic)

    # check we are doing a call over a custom dispatch function ----
    assert expr.func == "__call__"

    dispatcher = expr.args[0]
    assert isinstance(dispatcher, FuncArg)
    assert dispatcher.args[0] is array_ufunc    # could check .dispatch() method


def test_op_vars_slice(_):
    assert strip_symbolic(_.a[_.b:_.c]).op_vars() == {'a', 'b', 'c'}

# Truthiness should raise TypeError -------------------------------------------

import operator as op

@pytest.mark.parametrize('f, arg', [
    (op.contains, 'x'),
    (op.not_,None),
    (op.truth,None),
    (iter,None),
    ])
def test_siu_not_truth_value(_, f, arg):
    with pytest.raises(TypeError):
        f(_) if arg  is None else f(_, arg)
        f(_, arg)
    
def test_siu_not_truth_value_keywords(_):
    with pytest.raises(TypeError):
        _ and True

    with pytest.raises(TypeError):
        _ or True


# Explain  --------------------------------------------------------------------

from siuba.siu import explain

@pytest.mark.parametrize('code', [
    "-_",
    "+_",
    #abs,
    "~_",
    #complex,
    #int,
    #long,
    #float,
    #oct,
    #hex,
    #index,
    #round,
    #math.trunc
    #math.floor
    #math.ceil

    ])
def test_explain_unary(_, code):
    sym = eval(code, {'_': _})

    assert explain(sym) == code

@pytest.mark.parametrize('op', BINARY_OPS)
def test_explain_binary(_, op):
    left = "_ {op} 1".format(op = op)
    sym = eval(left, {'_': _})

    assert explain(sym) == left

    right = "1 {op} _".format(op = op)
    sym = eval(right, {'_': _})

    assert explain(sym) == right

@pytest.mark.parametrize('op', BINARY_OPS)
def test_explain_rhs_binary(_, op):
    right = "1 {op} _".format(op = op)

    sym = eval(right, {"_": _})

    assert explain(sym) == right

@pytest.mark.parametrize('op', BINARY_OPS)
def test_symbol_rhs_exec(_, op):
    if op == "@":
        pytest.skip("TODO: test with class where @ is implemented?")

    right = "1 {op} _".format(op = op)

    sym = eval(right, {'_': _})
    dst = eval(right, {'_': 2})

    assert sym(2) == dst

@pytest.mark.parametrize('expr', [
    "_['a']",
    "_.a",
    "_['a':'b']",
    "_['a':'b', 'c']",
    """_["'a'":'b']""",
    "_.a.mean(1,b = 2)",
    ])
def test_explain_other(_, expr):
    sym = eval(expr, {'_': _})
    assert explain(sym) == expr

@pytest.mark.parametrize('expr', [
    '_[slice(1,2)]',           # siu can't know concrete slice syntax
    '_["a"]',                  # siu can't know concrete quote syntax
    ])
def test_explain_failures(_, expr):
    sym = eval(expr, {'_': _})
    assert explain(sym) != expr


# Special Call subclasses =====================================================

# SliceOp ----

def test_sym_slice():
    from siuba.siu.calls import _SliceOpIndex

    _ = Symbolic()

    sym = _[_ < 1]
    meta, slice_ops = strip_symbolic(sym).args
    assert isinstance(meta, MetaArg)
    assert isinstance(slice_ops, Call)
    assert isinstance(slice_ops, SliceOp)       # abc metaclass
    assert slice_ops.__class__ is _SliceOpIndex


    indexer = slice_ops(1)
    assert indexer is False

def test_sym_slice_multiple():
    from siuba.siu.calls import _SliceOpExt

    _ = Symbolic()

    sym = _[_ < 1, :, :]

    meta, slice_ops = strip_symbolic(sym).args
    assert isinstance(meta, MetaArg)
    assert len(slice_ops.args) == 3
    assert isinstance(slice_ops.args[0], Call)
    assert isinstance(slice_ops, SliceOp)       # abc metaclass
    assert slice_ops.__class__ is _SliceOpExt

    indexer = slice_ops(1)
    assert indexer[0] is False
    assert indexer[1] == slice(None)
    
@pytest.mark.parametrize("expr, target", [
    ("_[1]", 1),
    ("_[1:]", slice(1, None)),
    ("_[:2]", slice(None, 2)),
    ("_[1:2]", slice(1, 2)),
    ("_[1:2:3]", slice(1, 2, 3)),
    ("_[slice(1,2,3)]", slice(1, 2, 3)),
    ("_[1, 'a']", (1, 'a')),
    ("_[(1, 'a')]", (1, 'a')),
    ("_[1, ]", (1,)),
    ("_[:, :, :]", (slice(None, None, None),)*3),
    ("_[_['a'], _['b']]", (1, 2)),
    ("_[_['a']:_['b']]", slice(1, 2)),
    ("_[_['a']:_['b'], :]", (slice(1, 2), slice(None))),
    ])
def test_slice_call_returns(_, expr, target):
    data = {'a': 1, 'b': 2}

    sym = eval(expr, {"_": _})
    index_op, slice_call = strip_symbolic(sym).args

    res = slice_call(data)
    assert res == target

# Node copying ================================================================

from siuba.siu.calls import Call, BinaryOp, SliceOp, MetaArg, FuncArg, DictCall
# Call
@pytest.mark.parametrize('node', [
    Call("__call__", lambda x, y = 2: x + y, 1, y = 2),
    BinaryOp("__add__", 1, 2),
    SliceOp("__siu_slice__", slice(0, 1)),
    SliceOp("__siu_slice__", (slice(0, 1), slice(2, 3))),
    MetaArg("_"),
    FuncArg("__custom_func__", lambda x: x),
    FuncArg(lambda x: x),
    DictCall("__call__", dict, {'a': 1, 'b': 2})
    ])
def test_node_copy(node):
    copy = node.copy()
    assert isinstance(copy, node.__class__)
    assert copy is not node
    assert copy.func == node.func
    assert copy.args == node.args
    assert copy.kwargs == node.kwargs

def test_dict_call_child_copy():
    bin_op = BinaryOp('__add__', 1, 2)
    call = DictCall("__call__", dict, {'a': bin_op, 'b': 2})
    copy = call.copy()

    assert isinstance(call.args[1]['a'], BinaryOp)
    assert bin_op is not copy.args[1]['a']


# Symbolic dispatch ===========================================================

from siuba.siu import symbolic_dispatch, Call, FuncArg

def test_FuncArg():
    f = lambda x: 1
    expr = FuncArg(f)

    assert expr(None) is f

def test_FuncArg_in_call():
    call = Call(
            '__call__',
            FuncArg(lambda x, y: x + y),
            1, y = 2
            )

    assert call(None) == 3



def test_symbolic_dispatch(_):
    @symbolic_dispatch
    def f(x, y = 2):
        return x + y

    # w/ simple Call
    call1 = f(strip_symbolic(_), 3)
    assert isinstance(call1, Call)
    assert call1(2) == 5

    # w/ simple Symbol
    sym2 = f(_, 3)
    assert isinstance(sym2, Symbolic)
    assert sym2(2) == 5

    # w/ complex Call
    sym3 = f(_['a'], 3)
    assert sym3({'a': 2}) == 5



# Call Tree Local =============================================================
from siuba.siu import CallTreeLocal

@pytest.fixture
def ctl():
    local = {'f_a': lambda self: self}
    yield CallTreeLocal(local, call_sub_attr = ('str', 'dt'))


def test_call_tree_local_sub_attr_method(_, ctl):
    # sub attr gets stripped w/ method call
    call = ctl.enter(strip_symbolic(_.str.f_a()))
    assert call('x') == 'x'

def test_call_tree_local_sub_attr_property(_, ctl):
    # sub attr gets stripped w/ property access
    call = ctl.enter(strip_symbolic(_.str.f_a))
    assert call('x') == 'x'

def test_call_tree_local_sub_attr_alone(_, ctl):
    # attr alone is treated like a normal getattr
    call = ctl.enter(strip_symbolic(_.str))
    assert call.func == "__getattr__"
    assert call.args[1] == "str"

def test_call_tree_local_sub_attr_method_missing(_, ctl):
    # subattr raises lookup errors (method)
    with pytest.raises(FunctionLookupError):
        ctl.enter(strip_symbolic(_.str.f_b()))

def test_call_tree_local_sub_attr_property_missing(_, ctl):
    # subattr raises lookup errors (property)
    with pytest.raises(FunctionLookupError):
        ctl.enter(strip_symbolic(_.str.f_b))

# symbolic dispatch and execution validation visitor ----
from siuba.siu import ExecutionValidatorVisitor

class SomeClass: pass

@pytest.fixture
def f_dispatch():
    @symbolic_dispatch
    def f(x):
        return 'default'

    @f.register(SomeClass)
    def _f_some_class(x):
        return 'some class'

    yield f


def test_call_tree_local_dispatch_cls_object(f_dispatch):
    ctl = ExecutionValidatorVisitor(
            dispatch_cls = object
            )

    call = Call("__call__", FuncArg(f_dispatch), MetaArg('_'))
    new_call = ctl.enter(call)
    assert new_call('na') == 'default'


def test_call_tree_local_dispatch_cls_subclass(f_dispatch):
    ctl = ExecutionValidatorVisitor(
            dispatch_cls = SomeClass
            )

    call = Call("__call__", FuncArg(f_dispatch), MetaArg('_'))
    new_call = ctl.enter(call)
    assert new_call('na') == 'some class'


# strict symbolic dispatch and call tree local ----

@pytest.fixture
def f_dispatch_strict():
    @symbolic_dispatch(cls = SomeClass)
    def f(x):
        return 'some class'

    yield f

def test_strict_dispatch_strict_default_fail(f_dispatch_strict):
    class Other(object): pass

    obj = Other()

    with pytest.raises(TypeError):
        f_dispatch_strict(obj)

def test_call_tree_local_dispatch_fail(f_dispatch_strict):
    ctl = CallTreeLocal(
            {'f_a': lambda self: self},
            dispatch_cls = object
            )

    call = Call("__call__", FuncArg(f_dispatch_strict), MetaArg('_'))

    # should be the default failure dispatch for object
    new_call = ctl.enter(call)
    with pytest.raises(TypeError):
        new_call('na')


# Codatavisitor ===============================================================
from siuba.siu.visitors import CodataVisitor

def test_codata_visitor_enclosed_call(_):
    class Alternative: pass

    @symbolic_dispatch
    def n(col):
        return f"n({col})"

    @symbolic_dispatch
    def mean(col):
        str_n = n(col)
        return f"sum({col}) / {str_n}"

    @n.register(Alternative)
    def _f(self, col):
        return f"n_alternative({col})"

    @mean.register(Alternative)
    def _f(self, col):
        str_n = n(self, col)
        return f"sum_alternative({col}) / {str_n}"

    codata = CodataVisitor(dispatch_cls = Alternative)

    expr = strip_symbolic(mean(_))
    new_expr = codata.visit(expr)

    assert new_expr("a") == "sum_alternative(a) / n_alternative(a)"

